import os
from pathlib import Path
from acm_hamburg_legacy import MoleculePreprocessorExtended
from rdkit import RDLogger


def preprocess_database(file) -> None:
    """Processes the molecules of a file and save them as a csv in ../preprocessedSmiles/outFilename.csv
    :param file: file with molecules to be processed.
    """
    fileName = os.path.basename(file).split(".")[0]
    outFilename = fileName + "_preprocessed.csv"
    outpath = Path(file).parent.parent / "preprocessedSmiles" / outFilename

    if Path(outpath).is_file():
        msg = f"File {outpath} already exists!"
        raise Exception(msg)

    smilesIDdict = convert_csv_to_smiles_dict(file)

    listOfSmiles = list(smilesIDdict.keys())
    rdLogger = RDLogger.logger()

    moleculesProcessed = (
        MoleculePreprocessorExtended.MoleculePreprocessorExtended.init_with_smiles(
            listOfSmiles,
        )
    )
    # set c++ log level of rdkit temporarily to ERROR so that csp does not clutter
    # stdout with logging
    rdLogger.setLevel(RDLogger.ERROR)
    moleculesProcessed.csp()  # use the chembl_structure_pipeline (standardizer and get_parent)
    rdLogger.setLevel(RDLogger.INFO)

    moleculesProcessed.element_filter()
    moleculesProcessed.check_molecules_validity()
    # rawsmiles:preprocessedSmiles dict
    preprocessedSmilesDict = moleculesProcessed.get_rawsmiles_smiles_dict()


    moleculesProcessed.canonalize_tautomer()
    # rawsmiles:tautomerizedSmiles dict
    tautoMoleculesDict = moleculesProcessed.get_rawsmiles_smiles_dict()


    logdir = Path(file).parent.parent / "log"
    if not logdir.exists():
        logdir.mkdir()
    logFileName = fileName + ".log"
    logFilepath = logdir / logFileName
    moleculesProcessed.write_log_file(logFilepath)

    # (rawSmiles, preprocessedSmiles, tautomerizedSmiles, ID)
    finalDict = mergeDict(
        tautoMoleculesDict,
        mergeDict(smilesIDdict, preprocessedSmilesDict),
    )

    save_dict_as_csv(
        finalDict,
        outpath,
        ["rawSmiles", "preprocessedSmiles", "id", "tautomerizedSmiles"],
    )


def mergeDict(dict1, dict2):
    """Merge dictionaries and keep values of common keys in list."""
    dict3 = {**dict1, **dict2}
    for key, value in dict3.items():
        if key in dict1 and key in dict2:
            dict3[key] = [value, dict1[key]]
    return dict3


def save_dict_as_csv(mergedDict, outputFile, headerList) -> None:
    """Saves the preprocessed molecules as smiles in a tab seperated csv.

    :param mergedDict: a rawSmiles: [[prepSmiles, id], canonSmiles] dictionary
    :param outputFile: filepath to the output file
    :param headerList: list with headers for the csv
    """
    with open(outputFile, "w", encoding="utf8") as csvFile:
        csvFile.write("\t".join(headerList) + "\n")
        for rawsmiles, values in mergedDict.items():
            csvFile.write(
                f"{rawsmiles}\t{values[0][0]}\t{values[0][1]}\t{values[1]}\n",
            )


def convert_csv_to_smiles_dict(file):
    """Expects a csv file with header line (which is skipped) and returns a rawsmiles:databaseID dict.

    :param file: a csv file, using 'space' as delimiter
    :return: a rawsmiles:databaseID dict
    """
    # rawsmiles:databaseID dict
    smilesIDdict = {}
    with open(file, encoding="utf8") as smilesIDfile:
        # skip first (header) line
        next(smilesIDfile)
        for line in smilesIDfile:
            entry = line.rstrip().split()
            smiles = entry[0]
            databaseID = entry[1]
            smilesIDdict[smiles] = databaseID
        return smilesIDdict



import click
from trialblazer import Trialblazer


@click.command()
@click.option("--input_file", help="Input File", type=str, required=True)
@click.option("--output_file", help="Output File", default="output.csv", type=str)
@click.option("--model_folder", help="Model Folder", default=None, type=str)
def main(input_file, output_file, model_folder) -> None:
    tb = Trialblazer(input_file=input_file, model_folder=model_folder)
    tb.run()
    tb.write(output_file=output_file)


@click.command()
@click.option("--url", help="Archive Model URL", type=str, required=False,default="https://zenodo.org/records/15783346/files/precalculated_data_for_trialblazer_model.tar.gz")
@click.option("--archive_type", help="Archive Type", type=str, required=False, default="tar.gz")
@click.option("--top_folder", help="Whether the archive has a top folder", type=str, required=False, default=True)
@click.option("--model_folder", help="Model Folder", default=None, type=str)
def download(url, archive_type,top_folder, model_folder) -> None:
    tb = Trialblazer(model_url=url, model_folder=model_folder,archive_type=archive_type,top_folder=top_folder)
    tb.download_model()


if __name__ == "__main__":
    main()



import numpy as np
import pandas as pd
from FPSim2 import FPSim2Engine
from rdkit import Chem


def target_features_preprocess(
    chembl_data,
    preprocessed_chembl_data,
    outputFolder,
    remove_multi_components=False,
):
    if remove_multi_components:
        removed_multicom_chembl = preprocessed_chembl_data[
            ~preprocessed_chembl_data.id.str.contains("x")
        ]
    else:
        removed_multicom_chembl = preprocessed_chembl_data
    chembl_data = chembl_data.rename(columns={"chembl_id": "id"})
    chembl_data["id"] = chembl_data["id"].astype(str)
    merged_data = removed_multicom_chembl.merge(
        chembl_data[["id", "target_id"]],
        how="left",
        on="id",
    )
    merged_data["MolWithoutStereo"] = merged_data["preprocessedSmiles"].apply(
        Chem.MolFromSmiles,
    )
    merged_data["SmilesWithoutStereo"] = merged_data["MolWithoutStereo"].apply(
        lambda x: Chem.MolToSmiles(x, isomericSmiles=False),
    )
    target_label_smpl = merged_data[["target_id", "SmilesWithoutStereo"]].dropna(
        subset=["target_id"],
    )
    target_id_list, preprocessed_target_unique_smiles, fpe = generate_h5file(
        target_label_smpl,
        outputFolder,
    )
    return target_id_list, preprocessed_target_unique_smiles, fpe


def tanimoto_similarity_calculation(
    fpe,
    query_dataset,
    split_data=False,
    num_splits=None,
):
    results_append = []
    my_smi_append = []
    if split_data:
        if not num_splits:
            msg = "num_splits must be provided when split_data is True."
            raise ValueError(
                msg,
            )
        gen = np.array_split(query_dataset, num_splits)
        for _index, batch in generator_yield_gen(gen):
            for my_smi in generator_yield_smiles(batch):
                results = fpe.similarity(my_smi, 0, n_workers=20)
                results_append.append(results)
                my_smi_append.append(my_smi)
    else:
        for my_smi in generator_yield_smiles(query_dataset):
            results = fpe.similarity(my_smi, 0, n_workers=31)
            results_append.append(results)
            my_smi_append.append(my_smi)
    return pd.DataFrame(
        {
            "my_smi": my_smi_append,
            "similarity_results": results_append,
        },
    )


def generate_h5file(preprocessed_target, outputFolder):
    target_id_list = list(preprocessed_target["target_id"].unique())
    target_id_list = [x for x in target_id_list if str(x) != "nan"]
    preprocessed_target_unique_smiles = (
        preprocessed_target.groupby("SmilesWithoutStereo")["target_id"]
        .apply(list)
        .reset_index()
    )
    preprocessed_target_unique_smiles["SmilesWithoutStereo"].tolist()

    fpe = FPSim2Engine(outputFolder)
    return target_id_list, preprocessed_target_unique_smiles, fpe


def generator_yield_gen(gen):
    yield from enumerate(gen)


def generator_yield_smiles(dataset):
    yield from dataset["SmilesForDropDu"]



import os
from pathlib import Path
from rdkit import Chem

"""
Script to split multi-component compounds into individual compounds. Use Chem.GetMolFrags(mol,asMols=True).
"""


def separate_multicomponents_test(inputDir) -> None:
    outdir = Path(inputDir).parent / "separate_multicom_GetMolFrags"
    if not outdir.exists():
        outdir.mkdir()
    count = 0
    counter = 0
    for filename in os.listdir(inputDir):
        outfilename = filename.split(".")[0] + "_noMulticom.csv"
        infile = inputDir / filename
        outfile = outdir / outfilename
        with open(infile) as inputSmilesCSV:
            with open(outfile, "w") as of:
                header = inputSmilesCSV.readline()
                of.write(header)
                for line in inputSmilesCSV:
                    entry = line.rstrip().split("\t")
                    if len(entry) == 4:
                        preprocessedSmile = entry[1]
                        name = entry[2]
                        tautomerizedSmiles = entry[3]

                        pre_mol = Chem.MolFromSmiles(preprocessedSmile)
                        tauto_mol = Chem.MolFromSmiles(tautomerizedSmiles)

                        molFrag_pre = Chem.GetMolFrags(pre_mol, asMols=True)
                        if tauto_mol is None:
                            pass
                        else:
                            molFrag_tauto = Chem.GetMolFrags(
                                tauto_mol,
                                asMols=True,
                            )

                        smilesList = [Chem.MolToSmiles(mol) for mol in molFrag_pre]
                        smilesList2 = [Chem.MolToSmiles(mols) for mols in molFrag_tauto]

                        if len(smilesList) == 1:
                            of.write(line)
                        else:
                            count += 1
                            for i in range(len(smilesList)):
                                new_name = str(name) + "x" + str(i)
                                of.write(
                                    entry[0]
                                    + "\t"
                                    + smilesList[i]
                                    + "\t"
                                    + new_name
                                    + "\t"
                                    + smilesList2[i]
                                    + "\n",
                                )
                    else:
                        counter += 1



from __future__ import annotations
import ast
import logging
import os
import pickle
import re
import tempfile
import zipfile
from pathlib import Path
import pandas as pd
import requests
from FPSim2 import FPSim2Engine
import tarfile

# Preprocess compounds
from rdkit import Chem
from rdkit.Chem import Descriptors
from rdkit.Chem import PandasTools
from .Dataset_preprocess.Preprocess_compound.Preprocess_FoodDB import (
    CheckOutputResult,
)

# these three functions are in folder Dataset_preprocess/Preprocess_compound
from .Dataset_preprocess.Preprocess_compound.Preprocess_FoodDB import (
    preprocess,
)
from .Dataset_preprocess.Preprocess_compound.remove_duplicate import (
    remove_duplicate_splitted_files,
)
from .Dataset_preprocess.Preprocess_compound.separate_multicomponents_test import (
    separate_multicomponents_test,
)

# these two functions are in folder Descriptor_calculation
from .Descriptor_calculation.process_similarity_results import (
    remove_tested_inactive_targets,
)
from .Descriptor_calculation.process_target_features import (
    add_morgan_fingerprints,
)
from .Descriptor_calculation.process_target_features import (
    process_target_features,
)
from .models import trialblazer_func
from .models import trialblazer_train

logger = logging.getLogger(__name__).addHandler(logging.NullHandler())


class Trialblazer:
    """Wrapper to load the model, the input smiles, set the different parameters and the methods, and then store the results."""

    def __init__(
        self,
        input_file: None | str = None,
        model_folder: None | str = None,
        M2FP_only: bool = False,
        threshold: float | None = None,
        k: int | None = None,
        remove_MultiComponent_cpd: bool = True,
        # features: None | list[str] = None,
        morgan_n_bits: int = 2048,
        model_url: str | None = "https://zenodo.org/records/15783346/files/precalculated_data_for_trialblazer_model.tar.gz",
        archive_type:str="tar.gz",
        top_folder:bool=True,
    ) -> None:
        """Create the triablazer object."""
        self.input_file = input_file
        self.k = k
        self.threshold = threshold
        self.remove_MultiComponent_cpd = remove_MultiComponent_cpd
        self.morgan_n_bits = morgan_n_bits
        if model_folder is None:
            self.model_folder = os.path.join(
                os.environ["HOME"], ".trialblazer", "models", "base_model",
            )
        else:
            self.model_folder = model_folder
        self.M2FP_only = M2FP_only
        if threshold is not None:
            self.threshold = threshold
        elif M2FP_only:
            self.threshold = 0.10338700489734302
        else:
            self.threshold = 0.06264154114736771
        if k is not None:
            self.k = k
        elif M2FP_only:
            self.k = 800
        else:
            self.k = 900
        self.model_url = model_url
        self.archive_type = archive_type
        self.top_folder=top_folder

    def import_smiles(self, smiles: list[str] | None = None) -> None:
        """Importing smiles either from the input file or from a list of smiles."""
        # if not hasattr(self, "smiles"):
        #     init_smiles = []
        # else:
        #     init_smiles = self.smiles
        # set_smiles = set(init_smiles)
        # self.smiles = init_smiles + [s for s in smiles if s not in set_smiles]
        if smiles is None:
            smiles = []
        smiles_df = pd.DataFrame([{"SMILES": s} for s in smiles])
        if not hasattr(self, "smiles"):
            self.smiles = smiles_df
        else:
            self.smiles = pd.concat(self.smiles, smiles_df)

    def import_smiles_file(
        self, smiles_file: str | None = None, force: bool = False,
    ) -> None:
        """Importing smiles either from the input file or from a list of smiles."""
        if not hasattr(self, "smiles") or force:
            if smiles_file is None:
                smiles_file = self.input_file
            if smiles_file is not None:
                smiles_df = pd.read_csv(smiles_file)

                if not hasattr(self, "smiles"):
                    self.smiles = smiles_df
                else:
                    self.smiles = pd.concat(self.smiles, smiles_df)
                # read_smiles = smiles_df["SMILES"].to_list()
                # self.import_smiles(read_smiles)

    def download_model(self,archive_type:str|None=None,top_folder:bool|None=None) -> None:
        if archive_type is None:
            archive_type=self.archive_type
        if top_folder is None:
            top_folder=self.top_folder
        if not os.path.exists(self.model_folder):
            os.makedirs(self.model_folder)
        if not os.path.exists(
            os.path.join(self.model_folder, "training_target_features.csv"),
        ):
            if self.model_url is not None:
                model_url = self.model_url
            elif "TRIALBLAZER_URL" in os.environ:
                model_url = os.environ["TRIALBLAZER_URL"]
            else:
                msg = "No specified value for model_url, aborting download."
                raise ValueError(msg)

            with tempfile.TemporaryDirectory() as tempdir:
                with requests.get(model_url, stream=True) as response:
                    response.raise_for_status()  # Raise an error for HTTP issues

                    # Save the archive file to a temporary location
                    archive_path = os.path.join(tempdir, f"temp_model.{archive_type}")
                    with open(archive_path, "wb") as temp_archive:
                        for chunk in response.iter_content(chunk_size=1024 * 1024):
                            if chunk:  # Filter out keep-alive chunks
                                temp_archive.write(chunk)

                if archive_type == 'zip':
                    with zipfile.ZipFile(archive_path, "r") as zip_file:
                        if top_folder:
                            all_members = zip_file.namelist()
                            members = [m for m in all_members if not m.startswith('.')]
                            top_level_folder = os.path.commonpath(members)

                            for member in members:
                                if member.endswith("/"):
                                    continue
                                member_path = os.path.relpath(member, top_level_folder)
                                target_path = os.path.join(self.model_folder, member_path)

                                print(f'Installing {target_path}')
                                os.makedirs(os.path.dirname(target_path), exist_ok=True)

                                with zip_file.open(member) as source, open(target_path, "wb") as target:
                                    target.write(source.read())
                        else:
                            zip_file.extractall(path=self.model_folder)

                elif archive_type == 'tar.gz':
                    with tarfile.open(archive_path, "r:gz") as tar:
                        if top_folder:
                            all_members = tar.getmembers()
                            members = [m for m in all_members if not m.name.startswith('.')]
                            top_level_folder = os.path.commonpath([m.name for m in members])

                            for member in members:
                                member_path = os.path.relpath(member.name, top_level_folder)
                                target_path = os.path.join(self.model_folder, member_path)
                                print(f'Installing {target_path}')

                                if member.isdir():
                                    os.makedirs(target_path, exist_ok=True)
                                else:
                                    os.makedirs(os.path.dirname(target_path), exist_ok=True)
                                    with tar.extractfile(member) as source, open(target_path, "wb") as target:
                                        target.write(source.read())
                        else:
                            tar.extractall(path=self.model_folder)

                else:
                    raise ValueError(f'Archive type not supported: {archive_type}')

    def run(self, force: bool = False) -> None:
        """Running model and storing results in self.result.
        If self.result already exists, recalculating only if "force" is specified.
        """
        if not hasattr(self, "result") or force:
            self.import_smiles_file(force=force)
            self.load_model()
            self.prepare_testset(force=force)
            self.run_model()

    def get_dataframe(self) -> pd.DataFrame:
        """Returns result as a dataframe."""
        if not hasattr(self, "result"):
            msg = "No result in Trialblazer object: Run the model first with the run() method"
            raise OSError(
                msg,
            )
        df = self.result.copy()

        PandasTools.AddMoleculeColumnToFrame(df, smilesCol="smi", molCol="mol")
        # df["ROMol"] = df["SMILES"].apply(Chem.MolFromSmiles)
        return df

    def write(
        self, output_file: str = "trialblazer_output.csv", sep: str = "|",
    ) -> None:
        """Write to file."""
        if not hasattr(self, "result"):
            msg = "No result in Trialblazer object: Run the model first with the run() method"
            raise OSError(
                msg,
            )
        self.result.set_index("id").sort_index().to_csv(
            output_file, index=True, sep=sep,
        )

    def run_model(self) -> None:
        """Once the model is loaded and the input data preprocessed, run the prediction."""
        if not hasattr(self, "model_data"):
            msg = "No model data in Trialblazer object: Load the model first with the load_model() method"
            raise OSError(
                msg,
            )
        self.result = trialblazer_func(
            classifier=self.classifier,
            selector=self.selector,
            test_set=self.test_set,
            threshold=self.threshold,
            training_fpe=self.model_data["training_data_fpe"],
            remove_MultiComponent_cpd=self.remove_MultiComponent_cpd,
            features=self.model_data["features"],
            training_set=self.model_data["training_target_features"],
        )

    @classmethod
    def preprocess(cls, moleculeCsv, out_folder=None, smiles_col="SMILES", id_col=None):
        """Preprocess the input data."""
        if out_folder is None:
            out_folder_obj = (
                tempfile.TemporaryDirectory()
            )  # os.path.join(test_folder_data, "..", "temp")
            out_folder = out_folder_obj.name

        """Step 1, preprocess compounds"""
        # refinedInputFile = "/data/local/Druglikness_prediction/external_test_set/approved_testset_final_withname.csv"

        # refinedInputFile = self.input_file
        # refinedInputFile =

        refinedOutputFolder = Path(out_folder)
        preprocess(
            moleculeCsv, refinedOutputFolder, smiles_col=smiles_col, id_col=id_col,
        )  # input: a dataframe with SMILES and ID, outpyut: a folder with the preprocessed files
        CheckOutputResult(
            refinedOutputFolder,
        )  # generate a csv file from log to check the output result

        """Step 2, separate multicomponents"""
        DB_processedDir = Path(out_folder) / "preprocessedSmiles"
        # the folder generated from step 1
        separate_multicomponents_test(
            DB_processedDir,
        )  # input: a folder with the preprocessed files, output: a folder with the separated multicomponents, for this step, I modified the script from anya

        """Step 3, remove duplicates"""
        refinedInputFolder = Path(out_folder) / "separate_multicom_GetMolFrags"
        # the folder generated from step 2
        refinedOutputFolder = refinedInputFolder.parent / "uniqueSmiles"
        if not refinedOutputFolder.exists():
            refinedOutputFolder.mkdir()
        remove_duplicate_splitted_files(
            refinedInputFolder, refinedOutputFolder, "approved_testset_final",
        )  # input: a folder with the separated multicomponents, output: a folder with the unique smiles files

        """Step 4, combine the unique smiles files, this step I haven't made it into a function"""
        unique_smiles_path = Path(out_folder) / "uniqueSmiles"

        df_list = []
        for filenames in os.listdir(unique_smiles_path):
            file = unique_smiles_path / filenames
            df = pd.read_csv(file, sep="\t", index_col=None, header=0)
            df_list.append(df)
        preprocessed_df = pd.concat(df_list, axis=0, ignore_index=True)

        """Step 5, remove stereochemical information from the preprocessed SMILES and filter the compounds to retain only small molecules"""
        preprocessed_df["Molecule"] = preprocessed_df["preprocessedSmiles"].apply(
            Chem.MolFromSmiles,
        )
        preprocessed_df["SmilesForDropDu"] = preprocessed_df["Molecule"].apply(
            Chem.MolToSmiles, isomericSmiles=False,
        )
        preprocessed_df["Molecule"] = preprocessed_df["SmilesForDropDu"].apply(
            Chem.MolFromSmiles,
        )
        preprocessed_df["mw"] = preprocessed_df.Molecule.apply(
            lambda mol: round(Descriptors.MolWt(mol), 3),
        )
        return preprocessed_df[preprocessed_df["mw"].between(150, 850)]

    def apply_tanimoto(self, preprocessed_df):
        """Step 7, calculate and process the Tanimoto similarity results, the query data is the preprocessed data from step 1-5, the output of this step is the target feature."""
        # load the preprocessed active and inactive targets from the ChEMBL database,
        # these targets are preprocessed through Step 1-5, but in the application it is not necessary to calculate it from the scratch

        temp_folder = tempfile.TemporaryDirectory()
        temp_folder2 = tempfile.TemporaryDirectory()
        output_path_temp_save_testdata_active = (
            temp_folder  # create a folder to save temporary files for the test data
        )
        output_path_temp_save_testdata_inactive = (
            temp_folder2  # create a folder to save temporary files for the test data
        )

        model_data = self.model_data
        (
            testset_active_binarize_list,
            testset_active_binarized_target_remain,
        ) = process_target_features(
            output_path_temp_save=output_path_temp_save_testdata_active,
            training_target_list=model_data[
                "training_target_list"
            ],  # only need to used the same targets as the training data
            fpe=model_data["active_fpe"],
            preprocessed_target_unique_smiles=model_data[
                "active_preprocessed_target_unique_smiles"
            ],
            query_data=preprocessed_df,
        )

        (
            testset_inactive_binarize_list,
            testset_inactive_binarized_target_remain,
        ) = process_target_features(
            output_path_temp_save=output_path_temp_save_testdata_inactive,
            training_target_list=model_data[
                "training_target_list"
            ],  # only need to used the same targets as the training data
            fpe=model_data["inactive_fpe"],
            preprocessed_target_unique_smiles=model_data[
                "inactive_preprocessed_target_unique_smiles"
            ],
            query_data=preprocessed_df,
        )
        testset_filtered_targets, testset_target_list = remove_tested_inactive_targets(
            testset_inactive_binarized_target_remain,
            testset_active_binarized_target_remain,
        )  # testset_filtered_targets is the target features I need for testset compounds

        testset_filtered_targets_id = testset_filtered_targets.merge(
            preprocessed_df[["SmilesForDropDu", "id"]],
            how="left",
            on="SmilesForDropDu",
        )

        """Step 8, calculate Morgan2 fingerprints for the training and test data"""
        morgan_cols = [f"morgan2_b{i}" for i in range(self.morgan_n_bits)]
        testset_filtered_targets_id = add_morgan_fingerprints(
            testset_filtered_targets_id, morgan_cols,
        )

        """Final step, employ the model"""
        # The input of Trialblazer is a dataframe of training featrues and the binary label of each compound, and the test set,
        # the output including a dataframe with the PrOCTOR socre and prediction results for each compound in test set, and the cloestest similairty between test compounds and training compounds
        return testset_filtered_targets_id

    def prepare_testset(self, out_folder=None, force=False) -> None:
        if not hasattr(self, "test_set") or force:
            preprocessed_df = self.preprocess(
                moleculeCsv=self.smiles, out_folder=out_folder,
            )
            test_set = self.apply_tanimoto(preprocessed_df=preprocessed_df)
            self.test_set = test_set
        else:
            pass

    def load_model(self, model_folder=None) -> None:
        """Load the model."""
        ##################### Model folder only
        """The following steps are used to calculate the descriptors for the compounds in dataset."""
        """Step 6, load the processed ChEMBL data and preprocessed training target features"""

        model_data = {}
        if model_folder is None:
            model_folder = self.model_folder
        inactive_preprocessed_target_unique_smiles = pd.read_csv(
            Path(model_folder)
            / "generated"
            / "target_preprocessed"
            / "inactive_target_preprocessed.csv",
            sep="|",
        )  # this preprocessed target smiles is precalculated by previous preprocessed steps base on chembl data, it doesn't need to be re-calculated.
        active_preprocessed_target_unique_smiles = pd.read_csv(
            Path(model_folder)
            / "generated"
            / "target_preprocessed"
            / "active_target_preprocessed.csv",
            sep="|",
        )  # this preprocessed target smiles is precalculated by previous preprocessed steps base on chembl data, it doesn't need to be re-calculated.
        active_preprocessed_target_unique_smiles[
            "target_id"
        ] = active_preprocessed_target_unique_smiles["target_id"].apply(
            ast.literal_eval,
        )  # this step is for converting the string type when I read the file from csv
        inactive_preprocessed_target_unique_smiles[
            "target_id"
        ] = inactive_preprocessed_target_unique_smiles["target_id"].apply(
            ast.literal_eval,
        )  # this step is for converting the string type when I read the file from csv

        # load the preprocessed training target features
        training_target_features = pd.read_csv(
            Path(model_folder) / "training_target_features.csv",
        )  # this training_target_features is calculated previously and don't need to be re-calculated

        # Load the generated active and inactive fpe (fingerprints engine)
        # These two fpe files are generated previously by using process_target_features function based on trainig data, and don't need to regenerated, I removed h5 files because if we have the fpe, then basically don't need the h5 file
        active_fpe = FPSim2Engine(
            Path(model_folder) / "generated" / "fingerprints" / "active_fpe.h5",
        )
        inactive_fpe = FPSim2Engine(
            Path(model_folder) / "generated" / "fingerprints" / "inactive_fpe.h5",
        )

        # load the preprocessed target features list, from the column names
        # with open(file_path, "r") as file:
        #     reader = csv.reader(file)
        #     training_target_list = [
        #         row[0] for row in reader
        #     ]  # get the list of traing_targets_list, number of targets: 777
        training_target_list = [
            c
            for c in training_target_features.columns
            if re.fullmatch(r"^CHEMBL[0-9]+$", c)
        ]
        morgan_cols = [f"morgan2_b{i}" for i in range(self.morgan_n_bits)]
        training_target_features = add_morgan_fingerprints(
            training_target_features, morgan_cols,
        )

        training_data_fpe = FPSim2Engine(
            Path(model_folder) / "generated" / "fingerprints" / "training_data_fpe.h5",
        )

        # M2FPs_PBFPs = morgan_cols + training_target_list
        model_data["y"] = training_target_features.Mark
        model_data["training_target_features"] = training_target_features
        model_data["active_fpe"] = active_fpe
        model_data["inactive_fpe"] = inactive_fpe
        model_data["training_target_list"] = training_target_list
        if self.M2FP_only:
            model_data["features"] = morgan_cols
        else:
            model_data["features"] = morgan_cols + training_target_list
        model_data["training_data_fpe"] = training_data_fpe
        model_data[
            "inactive_preprocessed_target_unique_smiles"
        ] = inactive_preprocessed_target_unique_smiles
        model_data[
            "active_preprocessed_target_unique_smiles"
        ] = active_preprocessed_target_unique_smiles
        self.model_data = model_data

        self.train_model()

    def save_classifier(self) -> None:
        classifier_path = Path(self.model_folder) / "generated" / "classifier.pkl"
        selector_path = Path(self.model_folder) / "generated" / "selector.pkl"
        if not os.path.exists(classifier_path):
            with open(classifier_path, "wb") as f:
                pickle.dump(self.classifier, file=f)
        if not os.path.exists(selector_path):
            with open(selector_path, "wb") as f:
                pickle.dump(self.selector, file=f)

    def load_classifier(self) -> None:
        classifier_path = Path(self.model_folder) / "generated" / "classifier.pkl"
        selector_path = Path(self.model_folder) / "generated" / "selector.pkl"
        with open(classifier_path, "rb") as f:
            self.classifier = pickle.load(f)
        with open(selector_path, "rb") as f:
            self.selector = pickle.load(f)

    def train_model(self, force=False, save=False, loadable=False) -> None:
        """Train the model if the file is not available. Save/Load is not advised: classifier depends on features chosen."""
        classifier_path = Path(self.model_folder) / "generated" / "classifier.pkl"
        if os.path.exists(classifier_path) and loadable:
            self.load_classifier()

        if not hasattr(self, "classifier") or force:
            self.classifier, self.selector = trialblazer_train(
                training_set=self.model_data["training_target_features"],
                y=self.model_data["y"],
                features=self.model_data["features"],
                k=self.k,
            )
            if save:
                self.save_classifier()



import os
import re
import zipfile
import pandas as pd

substring_1 = ["(T|t)oxic"]
substring_2 = [
    "(U|u)nrelated",
    "(N|n)ot (R|r)elated",
    "unknown",
    "(N|n)ot (S|s)tudy (R|r)elated",
    "non-related",
    "not due to",
    "other than",
    "(E|e)valuable",
    "Unable to complete",
    "no human safety concern",
    "concern for high systemic drug levels",
    "no participant had a dose limiting toxicity",
    "Enrol",
    "evalua",
    "Ongoing",
    "minimiz",
    "monitoring",
    "(R|r)eview",
    "rare",
    "potential",
    "non clinical toxicology",
    "gastrointestinal toxicity",
    "Toxicity assessment",
    "re-doing",
    "no excess of",
    "Unable to safely escalate",
    "check",
    "FDA",
    "concerns",
    "lack of response activity",
    "Investigators analyze the toxicity data",
]


def extract_filesFromZip(input_Path, dir_out) -> None:
    for filename in os.listdir(input_Path):
        if filename.endswith(".zip"):
            input_file = os.path.join(input_Path, filename)
            folder_name = input_file.split("/")[-1].split(".")[0]
            unzipFolder = dir_out / folder_name
            if not dir_out.exists():
                dir_out.mkdir()
            if not unzipFolder.exists():
                unzipFolder.mkdir()
            with zipfile.ZipFile(input_file) as zf:
                zf.extract("studies.txt", unzipFolder)
                zf.extract("interventions.txt", unzipFolder)
                zf.extract("drop_withdrawals.txt", unzipFolder)


def map_names_intervention(
    Agg_name,
    toxicity_negative,
    toxicity_negative_excluded,
):
    # Use "interventions set" to map the name of drugs through nct_id
    lst = [";", "and"]
    OneDrug = Agg_name[~Agg_name["name"].str.contains("|".join(lst))]
    OneDrug = OneDrug.rename(columns={"name": "Name"})
    toxicity_negative_sim = toxicity_negative[
        ["nct_id", "reason", "why_stopped", "phase", "overall_status"]
    ]
    toxicity_negative_excluded_sim = toxicity_negative_excluded[
        ["nct_id", "reason", "why_stopped", "phase", "overall_status"]
    ]
    # match the name using nct_id
    toxic_onedrug = OneDrug.merge(
        toxicity_negative_sim,
        how="right",
        on="nct_id",
    )
    benign_onedrug = OneDrug.merge(
        toxicity_negative_excluded_sim,
        how="right",
        on="nct_id",
    )

    return toxic_onedrug, benign_onedrug


def keywords_filtering(df, category):
    contains_sub1 = df[category].str.contains("|".join(substring_1))
    contains_sub2 = df[category].str.contains("|".join(substring_2))
    selected_toxicity = df[contains_sub1 & ~contains_sub2]
    non_toxicity = df[~contains_sub1 | contains_sub2]
    return selected_toxicity, non_toxicity


def toxicity_set(studies_set, drop_withdrawals_set):
    # dropna in the "why_stopped" and "reason" categories
    drop_withdrawals_dropna = drop_withdrawals_set.dropna(subset=["reason"])
    studies_simplify_dropna = studies_set.dropna(subset=["why_stopped"])

    studies_simplify_dropna_sim = studies_simplify_dropna[
        ["nct_id", "why_stopped", "phase", "overall_status"]
    ]
    drop_withdrawals_dropna_sim = drop_withdrawals_dropna[["nct_id", "reason"]]

    # nan_set, prepare for the benign set
    drop_withdrawals_set_nan = drop_withdrawals_set[
        drop_withdrawals_set["reason"].isna()
    ]
    studies_set_nan = studies_set[studies_set["why_stopped"].isna()]
    merged_two_nan_set = drop_withdrawals_set_nan.merge(
        studies_set_nan,
        how="outer",
        on="nct_id",
    )
    merged_two_nan_set_sim = merged_two_nan_set[
        ["nct_id", "reason", "why_stopped", "phase", "overall_status"]
    ]

    # filter toxicity description in "why_stopped" and "reason" categories
    selected_toxicity_why_stopped, non_toxicity_why_stopped = keywords_filtering(
        studies_simplify_dropna_sim, "why_stopped",
    )
    selected_toxicity_reason, non_toxicity_reason = keywords_filtering(
        drop_withdrawals_dropna_sim,
        "reason",
    )
    toxicity_negative = selected_toxicity_reason.merge(
        selected_toxicity_why_stopped,
        how="outer",
        on="nct_id",
    )

    # filter the non-toxic set (benign_ori)
    toxicity_negative_excluded = non_toxicity_reason.merge(
        non_toxicity_why_stopped,
        how="outer",
        on="nct_id",
    )
    toxicity_negative_excluded_inclnan = merged_two_nan_set_sim.append(
        toxicity_negative_excluded,
        ignore_index=True,
    )
    return toxicity_negative, toxicity_negative_excluded_inclnan


def check_name(interventions_set):
    interventions_Drug = interventions_set.loc[
        interventions_set.intervention_type == "Drug"
    ]
    Name_exclude_placebo = interventions_Drug[
        interventions_Drug.name.str.contains("(P|p)lacebo") is False
    ]
    Name_exclude_placebo_du = Name_exclude_placebo.drop_duplicates(
        subset=["nct_id", "name"],
        keep="last",
    )  # deduplicates the drugs in the same trial but using different doses
    Agg_name = (
        Name_exclude_placebo_du.groupby(["nct_id"])["name"]
        .agg(lambda x: "; ".join(x.astype(str)))
        .reset_index()
    )
    return Agg_name, Name_exclude_placebo_du


def preprocess_aact(unzipFolder_path):
    toxicity_negative_append = []
    Agg_name_append = []
    toxicity_negative_excluded_append = []
    intervention_set_ori_append = []
    for filename in os.listdir(unzipFolder_path):
        match = re.search(r"\d{6}", filename)
        if not filename.endswith(".zip") and match:
            input_file = os.path.join(unzipFolder_path, filename)

            studies_str_path = input_file + "/" + "studies" + ".txt"
            interventions_str_path = input_file + "/" + "interventions" + ".txt"
            drop_withdrawals_str_path = input_file + "/" + "drop_withdrawals" + ".txt"

            studies_set = pd.read_csv(studies_str_path, sep="|")
            interventions_set = pd.read_csv(interventions_str_path, sep="|")
            drop_withdrawals_set = pd.read_csv(
                drop_withdrawals_str_path,
                sep="|",
            )

            Agg_name, intervention_set_ori = check_name(interventions_set)
            toxicity_negative, toxicity_negative_excluded_inclnan = toxicity_set(
                studies_set, drop_withdrawals_set,
            )

            # Append the results in different files
            toxicity_negative_append.append(toxicity_negative)
            toxicity_negative_excluded_append.append(
                toxicity_negative_excluded_inclnan,
            )
            Agg_name_append.append(Agg_name)
            intervention_set_ori_append.append(intervention_set_ori)

    toxicity_negative_set_total = pd.concat(toxicity_negative_append, axis=0)
    toxicity_negative_excluded_append_total = pd.concat(
        toxicity_negative_excluded_append,
        axis=0,
    )
    Agg_name_append_total = pd.concat(Agg_name_append, axis=0)
    intervention_set_ori_total = pd.concat(intervention_set_ori_append, axis=0)
    (
        toxicity_negative_onedrug,
        toxicity_negative_exclude_onedrug,
    ) = map_names_intervention(
        Agg_name_append_total,
        intervention_set_ori_total,
        toxicity_negative_set_total,
        toxicity_negative_excluded_append_total,
    )
    return (
        Agg_name_append_total,
        toxicity_negative_onedrug,
        toxicity_negative_exclude_onedrug,
        intervention_set_ori_total,
    )


def prepocess_wholeset(count_path) -> None:
    toxic_append = []
    Agg_name_append = []
    benign_onedrug_append = []
    intervention_set_ori_append = []
    for filename in os.listdir(count_path):
        match = re.search(r"\d{4}", filename)
        if not filename.endswith(".zip") and match:
            input_file = os.path.join(count_path, filename)
            (
                Agg_name,
                toxic_onedrug,
                benign_onedrug,
                intervention_set_ori,
            ) = preprocess_aact(input_file)

            toxic_append.append(toxic_onedrug)
            benign_onedrug_append.append(benign_onedrug)
            Agg_name_append.append(Agg_name)
            intervention_set_ori_append.append(intervention_set_ori)

    toxic_append_total = pd.concat(toxic_append, axis=0)
    Agg_name_append_total = pd.concat(Agg_name_append, axis=0)
    benign_append_total = pd.concat(benign_onedrug_append, axis=0)
    intervention_set_ori_append_total = pd.concat(
        intervention_set_ori_append,
        axis=0,
    )

    toxic_append_total.nct_id.unique().size
    Agg_name_append_total.nct_id.unique().size
    intervention_set_ori_append_total.nct_id.unique().size

    check_other_phase = count_path + "/" + "check_other_phase_1"
    if not os.path.exists(check_other_phase):
        os.makedirs(check_other_phase)

    Merge_name_path = check_other_phase + "/" + "toxicity_multi_drugs" + ".csv"
    toxicity_negative_excluded_path = (
        check_other_phase + "/" + "toxicity_multidrugs_excluded" + ".csv"
    )
    Agg_name_append_total_append_total_path = (
        check_other_phase + "/" + "Agg_name" + ".csv"
    )
    intervention_set_ori_append_total_path = (
        check_other_phase + "/" + "intervention_set_ori" + ".csv"
    )


    toxic_append_total.to_csv(Merge_name_path, sep="|")
    benign_append_total.to_csv(toxicity_negative_excluded_path, sep="|")
    Agg_name_append_total.to_csv(
        Agg_name_append_total_append_total_path,
        sep="|",
    )
    intervention_set_ori_append_total.to_csv(
        intervention_set_ori_append_total_path,
        sep="|",
    )



import argparse
import os
from pathlib import Path

"""
Helper script to split large files, used before preprocess_molecules.py
"""


def split_large_files(file, filesize) -> None:
    """Splits a large file with molecules and ID into smaller files of the given file size and saves them in
    ../smiles_input/databse_n.smi.

    :param file: large file
    :param filesize: number of lines in the smaller files
    """
    outputFolder = Path(file).parent.parent / "smiles_input"
    if not outputFolder.exists():
        outputFolder.mkdir()

    smallfile = None
    with open(file, encoding="utf8") as bigfile:
        for lineno, line in enumerate(bigfile):
            if lineno % filesize == 0:
                if smallfile:
                    smallfile.close()
                fileName = os.path.basename(file).split(".")[0]
                outFilename = fileName + "_" + str(lineno // filesize) + ".smi"
                outpath = outputFolder / outFilename
                smallfile = open(outpath, "w", encoding="utf-8")
                if lineno != 0:
                    smallfile.write("smiles ID\n")
            smallfile.write(line)
        if smallfile:
            smallfile.close()


if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument(
        "-i",
        "--inputfilepath",
        type=str,
        help="Specify input filepath with files to split.",
        default="/data/local/data_food/check_structure/Preprocess",
    )
    p.add_argument(
        "-n",
        "--splitnumber",
        type=int,
        help="Number of molecules in one file.",
        default=3000,
    )
    args = p.parse_args()

    for file in os.listdir(args.inputfilepath):
        if file.endswith(".smi"):
            split_large_files(
                os.path.join(args.inputfilepath, file),
                args.splitnumber,
            )



import os
import sqlite3
import tarfile
import tempfile
import numpy as np
import pandas as pd
import requests
from FPSim2.io import create_db_file
from rdkit import Chem
from .trialblazer import Trialblazer


def label_str_to_int(x):
    if x == "inactive":
        return 0
    if x == "active":
        return 1
    return np.nan


class TrialTrainer:
    """Used to train a model from scratch.
    Needed:
     - A version of chembl
     - A training set file (curated)
     - Optional: an extra test set (typically all benign).
    """

    chembl_query = """
WITH valid_data AS (
    SELECT *
    FROM activities
    WHERE standard_type IN ('Kd', 'Potency', 'AC50', 'IC50', 'Ki', 'EC50')
    AND standard_relation == '='
    AND standard_units == 'nM'
    AND potential_duplicate == 0
    AND (data_validity_comment IS NULL OR data_validity_comment == 'Manually validated')
)
SELECT assays.chembl_id AS assay_id,
       assay_type,
       target_dictionary.pref_name,
       standard_type,
       molecule_dictionary.chembl_id AS molecule_id,
       target_dictionary.chembl_id AS target_id,
       canonical_smiles,
       standard_value,
       pchembl_value
FROM valid_data
LEFT JOIN compound_structures USING (molregno)
LEFT JOIN assays USING (assay_id)
LEFT JOIN target_dictionary USING (tid)
LEFT JOIN molecule_dictionary USING (molregno)
INNER JOIN (
    SELECT assay_id,
           standard_type
    FROM valid_data
    GROUP BY assay_id, standard_type
    ) USING (
        assay_id,
        standard_type
        )
    WHERE confidence_score IN (7, 8, 9)
"""

    def __init__(
        self,
        chembl_version=34,
        training_set=None,
        model_folder=None,
        extra_test_set=None,
        chembl_folder=None,
        archive_folder=None,
        inactive_threshold=20000,
        active_threshold=10000,
        size_limit=None,
        morgan_nbits=2048,
    ):
        self.chembl_version = chembl_version
        if chembl_folder is None:
            self.chembl_folder = os.path.join(
                os.environ["HOME"], ".trialblazer", "chembl",
            )
        else:
            self.chembl_folder = chembl_folder
        if model_folder is None:
            self.model_folder = os.path.join(
                os.path.dirname(__file__),
                "data",
                "base_model",
            )
        else:
            self.model_folder = model_folder
        if training_set is None:
            self.training_set = os.path.join(
                self.model_folder,
                "training_target_features.csv",
            )
        else:
            self.training_set = training_set
        self.extra_test_set = extra_test_set
        self.archive_folder = archive_folder
        self.inactive_threshold = inactive_threshold
        self.active_threshold = active_threshold
        self.size_limit = size_limit
        self.morgan_nbits = morgan_nbits

    def chembl_download(self, version=None) -> None:
        """Downloading and decompressing the archive if the database file does not exist."""
        if not os.path.exists(self.chembl_folder):
            os.makedirs(self.chembl_folder)
        filepath = os.path.join(
            self.chembl_folder, f"chembl_{self.chembl_version}.sqlite",
        )
        if not os.path.exists(filepath):
            filename = f"chembl_{self.chembl_version}_sqlite.tar.gz"
            url = (
                f"https://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_{self.chembl_version}/{filename}",
            )
            with tempfile.TemporaryDirectory() as tmpdir:
                if self.archive_folder is None:
                    output_path = os.path.join(tmpdir, filename)
                else:
                    output_path = os.path.join(
                        self.archive_folder,
                        filename,
                    )
                if not os.path.exists(output_path):
                    with requests.get(url, stream=True) as response:
                        response.raise_for_status()  # Raise an error for bad status codes (e.g., 404, 500)
                        with open(output_path, "wb") as file:
                            for chunk in response.iter_content(chunk_size=1024 * 1024):
                                if chunk:
                                    file.write(chunk)
                target_file = f"chembl_{self.chembl_version}/chembl_{self.chembl_version}_sqlite/chembl_{self.chembl_version}.db"
                with tarfile.open(output_path, "r:gz") as tar:
                    # Check if the target file exists in the archive
                    if target_file in tar.getnames():
                        # Extract the specific file to the desired location
                        with tar.extractfile(target_file) as file:
                            buffer_size = 50 * 1024 * 1024
                            with open(filepath, "wb") as output_file:
                                while True:
                                    chunk = file.read(buffer_size)
                                    if not chunk:  # Stop when no more data is available
                                        break
                                    output_file.write(chunk)
                    else:
                        for _f in tar.getnames():
                            pass

    def process_activity(self, con=None) -> None:
        if con is None:
            with sqlite3.connect(
                os.path.join(self.chembl_folder, f"chembl_{self.chembl_version}.sqlite"),
            ) as con:
                df = pd.read_sql(self.chembl_query, con=con, chunksize=self.size_limit)
                if self.size_limit is not None:
                    df = next(df)
        else:
            df = pd.read_sql(self.chembl_query, con=con, chunksize=self.size_limit)
            if self.size_limit is not None:
                df = next(df)
        df = df.dropna()
        # remove stereochemistry information and using median activity value as representative activity value for the compounds
        df["mol"] = df["canonical_smiles"].apply(Chem.MolFromSmiles)
        df["SMILES_withoutStereoChem"] = df.mol.apply(
            Chem.MolToSmiles, isomericSmiles=False,
        )
        df_grouped_median = (
            df.groupby(["SMILES_withoutStereoChem", "target_id"])["standard_value"]
            .median()
            .reset_index()
        )

        # get the median number of activity value for the same target, same compounds that were tested in different assays
        df_grouped_median = (
            df.groupby(["SMILES_withoutStereoChem", "target_id"])["standard_value"]
            .median()
            .reset_index()
        )

        df_grouped_median["LABEL"] = df_grouped_median["standard_value"].apply(
            self.activity_filter,
        )
        df_grouped_median = df_grouped_median.rename(
            columns={"standard_value": "standard_value_median"},
        )
        df_grouped_median = df_grouped_median.drop_duplicates()

        df_grouped_median["LABEL"] = df_grouped_median["LABEL"].map(label_str_to_int)
        df_grouped_median_active = df_grouped_median[df_grouped_median.LABEL == 1]
        df_grouped_median_inactive = df_grouped_median[df_grouped_median.LABEL == 0]

        # here the preprocess means the steps 1-5 in model Trialblazer
        self.active_target_preprocessed = self.preprocess(df_grouped_median_active)
        self.inactive_target_preprocessed = self.preprocess(df_grouped_median_inactive)

    def write_target_preprocessed(self, output_folder=None, force=False) -> None:
        if output_folder is None:
            output_folder = os.path.join(
                self.model_folder, "generated", "target_preprocessed",
            )
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        filepath = os.path.join(output_folder, "active_target_preprocessed.csv")
        if force or not os.path.exists(filepath):
            self.active_target_preprocessed.to_csv(
                filepath,
                index=False,
                sep="|",
            )
        filepath = os.path.join(output_folder, "inactive_target_preprocessed.csv")
        if force or not os.path.exists(filepath):
            self.inactive_target_preprocessed.to_csv(
                filepath,
                index=False,
                sep="|",
            )

    def preprocess(self, moleculeCsv):
        df = Trialblazer.preprocess(
            moleculeCsv=moleculeCsv,
            smiles_col="SMILES_withoutStereoChem",
            id_col="target_id",
        )
        df = df[["SmilesForDropDu", "id"]].rename(
            columns={"SmilesForDropDu": "SmilesWithoutStereo", "id": "target_id"},
        )
        df["target_id"] = df["target_id"].apply(lambda x: [x])
        return df

    def activity_filter(self, x):
        if x >= self.inactive_threshold:
            return "inactive"
        if x < self.active_threshold:
            return "active"
        return np.nan

    def write_h5(self, smiles, filename, output_folder=None, force=False) -> None:
        """Writing h5 fingerprints database from FPSim2."""
        if output_folder is None:
            output_folder = os.path.join(self.model_folder, "generated", "fingerprints")
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        output_file = os.path.join(output_folder, filename)
        if force or not os.path.exists(output_file):
            smiles_gen = ((s, i) for i, s in enumerate(smiles))
            create_db_file(
                mols_source=smiles_gen,
                filename=output_file,
                mol_format="smiles",
                fp_type="Morgan",
                fp_params={"radius": 2, "fpSize": self.morgan_nbits},
            )
        else:
            pass

    def build_model_data(self, con=None, cleanup=True) -> None:
        preprocessed_folder = os.path.join(
            self.model_folder, "generated", "preprocessed",
        )
        fpe_folder = os.path.join(self.model_folder, "generated", "fingerprints")
        if not os.path.exists(
            os.path.join(fpe_folder, "active_fpe.h5"),
        ) or not os.path.exists(os.path.join(fpe_folder, "inactive_fpe.h5")):
            if os.path.exists(
                os.path.join(preprocessed_folder, "active_target_preprocessed.csv"),
            ):
                self.active_target_preprocessed = pd.read_csv(
                    os.path.join(preprocessed_folder, "active_target_preprocessed.csv"),
                    sep="|",
                )
            if os.path.exists(
                os.path.join(preprocessed_folder, "inactive_target_preprocessed.csv"),
            ):
                self.inactive_target_preprocessed = pd.read_csv(
                    os.path.join(
                        preprocessed_folder, "inactive_target_preprocessed.csv",
                    ),
                    sep="|",
                )
            if not hasattr(self, "active_target_preprocessed") or not hasattr(
                self, "inactive_target_preprocessed",
            ):
                if con is None:
                    self.chembl_download()
                self.process_activity(con=con)
                self.write_target_preprocessed()
            else:
                pass
            self.write_h5(
                smiles=self.active_target_preprocessed["SmilesWithoutStereo"],
                filename="active_fpe.h5",
            )
            self.write_h5(
                smiles=self.inactive_target_preprocessed["SmilesWithoutStereo"],
                filename="inactive_fpe.h5",
            )
        else:
            pass
        if not os.path.exists(os.path.join(fpe_folder, "training_data_fpe.h5")):
            self.load_training_data()
            self.write_h5(
                smiles=self.training_data["SmilesForDropDu"],
                filename="training_data_fpe.h5",
            )

        else:
            pass
        if cleanup:
            self.cleanup()

    def load_training_data(self, sep=",") -> None:
        if not hasattr(self, "training_data"):
            self.training_data = pd.read_csv(self.training_set, sep=sep)

    def cleanup(self) -> None:
        for a in (
            "inactive_target_preprocessed",
            "active_target_preprocessed",
            "training_data",
        ):
            if hasattr(self, a):
                delattr(self, a)



from rdkit import Chem
from rdkit.Chem import AllChem
from .process_similarity_results import binarize_similarity_value
from .process_similarity_results import organize_similarity_results
from .process_similarity_results import separate_similarity_results
from .similarity_based_target_calculation import (
    tanimoto_similarity_calculation,
)
from .similarity_based_target_calculation import target_features_preprocess


def process_target_features(
    output_path_temp_save,
    similarity_threshold=0.75,
    training_data=False,
    assay_median_value=None,
    target_prepocessed=None,
    output_h5=None,
    training_target_list=None,
    fpe=None,
    preprocessed_target_unique_smiles=None,
    query_data=None,
):
    if training_data:
        (
            target_id_list,
            preprocessed_target_unique_smiles,
            fpe,
        ) = target_features_preprocess(
            assay_median_value,
            target_prepocessed,
            output_h5,
        )
        data_for_similarity = query_data
        target_ids = target_id_list
    else:
        if (
            training_target_list is None
            or fpe is None
            or preprocessed_target_unique_smiles is None
            or query_data is None
        ):
            msg = "For test mode, training_target_list, fpe, preprocessed_target_unique_smiles, and query_data are required"
            raise ValueError(
                msg,
            )
        data_for_similarity = query_data
        target_ids = training_target_list

    results_whole = tanimoto_similarity_calculation(
        fpe,
        data_for_similarity,
    )  # kernal crush

    separate_similarity_results(
        results_whole,
        target_ids,
        preprocessed_target_unique_smiles,
        output_path_temp_save,
        training_data=training_data,
    )  # kernal crush

    features, target_list = organize_similarity_results(
        target_ids,
        output_path_temp_save,
        training_data=training_data,
    )

    features = features.rename(columns={"smi": "SmilesForDropDu"})

    if not training_data:
        pass
    binarize_list, binarized_target_remain = binarize_similarity_value(
        features,
        target_list,
        similarity_threshold,
    )

    if training_data:
        return (
            binarize_list,
            binarized_target_remain,
            preprocessed_target_unique_smiles,
            fpe,
        )
    return binarize_list, binarized_target_remain


def add_morgan_fingerprints(df, morgan_cols, smiles_column="SmilesForDropDu"):
    result_df = df.copy()
    result_df["Molecule"] = result_df[smiles_column].apply(Chem.MolFromSmiles)
    result_df["fp"] = result_df.Molecule.apply(get_morgan2)
    result_df[morgan_cols] = result_df["fp"].to_list()
    return result_df


def get_morgan2(mol):
    return list(AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=2048))



"""Documentation about trialblazer."""

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Huanni Zhang"
__email__ = "huanni.zhang@univie.ac.at"
__version__ = "0.1.0"

from .trialblazer import Trialblazer
from .trialtrainer import TrialTrainer



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from rdkit.Chem import AllChem
from rdkit.Chem import Descriptors
from rdkit.ML.Descriptors import MoleculeDescriptors
from scipy.stats import bernoulli
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import f_classif
from sklearn.metrics import ConfusionMatrixDisplay
from sklearn.metrics import RocCurveDisplay
from sklearn.metrics import auc
from sklearn.metrics import balanced_accuracy_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import StratifiedKFold
from sklearn.neural_network import MLPClassifier
from tqdm import tqdm


def get_morgan2(mol):
    return list(AllChem.GetMorganFingerprintAsBitVect(mol,2,nBits=2048))


def compute_2Drdkit(mol):
    calc = MoleculeDescriptors.MolecularDescriptorCalculator([x[0] for x in Descriptors._descList])
    ds = calc.CalcDescriptors(mol)
    return list(ds)

def MLP_simulation_test(X_new,y):
    cv = StratifiedKFold(n_splits=10)
    MCCs = []
    for _i, (train, test) in enumerate(cv.split(X_new, y)):
        p_estimate = np.mean(y[train])
        n_test_samples = len(X_new[test])
        y_random = bernoulli.rvs(p_estimate, size=n_test_samples)
        mcc = matthews_corrcoef(y[test],y_random)
        MCCs.append(mcc)
    return np.mean(MCCs)


def MLP_decision_threshold_optimization(X, y, opt_num_feature):
    selector = SelectKBest(f_classif, k=opt_num_feature)
    X_new = selector.fit_transform(X, y)
    cv = StratifiedKFold(n_splits=10)
    classifier = MLPClassifier(
        hidden_layer_sizes=(10,),
        random_state=42,
        learning_rate_init=0.0001,
        max_iter=600,
    )
    opt_threshold_ap = []
    for _i, (train, test) in enumerate(cv.split(X_new, y)):
        classifier.fit(X_new[train], y[train])
        y_prob = classifier.predict_proba(X_new[test])[:, 1]
        fpr, tpr, thresholds = metrics.roc_curve(y[test], y_prob)
        opt_id = np.argmax(tpr - fpr)
        opt_threshold = thresholds[opt_id]
        opt_threshold_ap.append(opt_threshold)
    opt_threshold_mean = np.mean(opt_threshold_ap)
    return opt_threshold_ap, opt_threshold_mean


def MLP_cv(X, y, opt_num_feature="all", threshold=None) -> None:
    selector = SelectKBest(f_classif, k=opt_num_feature)
    X_new = selector.fit_transform(X, y)
    cv = StratifiedKFold(n_splits=10)
    if opt_num_feature == "all" and threshold is None:
        classifier = MLPClassifier(random_state=42)
    else:
        classifier = MLPClassifier(
            hidden_layer_sizes=(10,),
            random_state=42,
            learning_rate_init=0.0001,
            max_iter=600,
        )
    tprs, aucs, MCCs, cms, baccs, recalls, precisions = (
        [],
        [],
        [],
        [],
        [],
        [],
        [],
    )
    mean_fpr = np.linspace(0, 1, 100)
    fig, ax = plt.subplots()
    fig1, ax1 = plt.subplots()

    for i, (train, test) in enumerate(cv.split(X_new, y)):
        classifier.fit(X_new[train], y[train])
        loss_curve = classifier.loss_curve_
        iter = classifier.n_iter_

        y_prob = classifier.predict_proba(X_new[test])[:, 1]

        if threshold is None:  # Default MLP
            y_pred_opt = classifier.predict(X_new[test])
        else:  # Optimized MLP with custom threshold
            y_pred_opt = (y_prob >= threshold).astype(int)

        mcc = matthews_corrcoef(y[test], y_pred_opt)
        bacc = balanced_accuracy_score(y[test], y_pred_opt)
        cm = confusion_matrix(y[test], y_pred_opt, labels=classifier.classes_)
        rec = recall_score(y[test], y_pred_opt)
        prec = precision_score(y[test], y_pred_opt)

        MCCs.append(mcc)
        baccs.append(bacc)
        cms.append(cm)
        recalls.append(rec)
        precisions.append(prec)
        # ROC curve
        viz = RocCurveDisplay.from_estimator(
            classifier,
            X_new[test],
            y[test],
            name=f"ROC fold {i}",
            alpha=0.3,
            lw=1,
            ax=ax,
        )
        # loss function curve
        iteration = np.linspace(0, iter, iter)
        ax1.plot(
            iteration,
            loss_curve,
            color="mediumslateblue",
            lw=1,
            alpha=0.5,
        )
        interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
        interp_tpr[0] = 0.0
        tprs.append(interp_tpr)
        aucs.append(viz.roc_auc)
        MCCs.append(mcc)
        cms.append(cm)
        baccs.append(bacc)
        recalls.append(rec)
        precisions.append(prec)

    mean_tpr = np.mean(tprs, axis=0)
    mean_tpr[-1] = 1.0
    mean_auc = auc(mean_fpr, mean_tpr)
    std_auc = np.std(aucs)
    np.mean(MCCs)
    np.std(MCCs)
    stacked_matrices = np.stack(cms, axis=0)
    mean_cms = np.mean(stacked_matrices, axis=0)
    mean_cms = np.round(mean_cms).astype(
        int,
    )  # get the integer of the mean number
    np.mean(baccs)
    np.std(baccs)
    np.mean(recalls)
    np.std(recalls)
    np.mean(precisions)
    np.std(precisions)

    ax1.set(
        xlabel="Iteration",
        ylabel="Loss",
        title="Loss function curve",
    )
    ax.plot(
        [0, 1],
        [0, 1],
        linestyle="--",
        lw=2,
        color="r",
        label="Chance",
        alpha=0.8,
    )
    ax.plot(
        mean_fpr,
        mean_tpr,
        color="b",
        label=rf"Mean ROC (AUC = {mean_auc:0.2f} $\pm$ {std_auc:0.2f})",
        lw=2,
        alpha=0.8,
    )
    std_tpr = np.std(tprs, axis=0)
    tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
    tprs_lower = np.maximum(mean_tpr - std_tpr, 0)

    ax.fill_between(
        mean_fpr,
        tprs_lower,
        tprs_upper,
        color="grey",
        alpha=0.2,
        label=r"$\pm$ 1 std. dev.",
    )
    ax.set(
        xlim=[-0.05, 1.05],
        ylim=[-0.05, 1.05],
        title="ROC Curve",
    )
    ax.legend(loc="lower right")
    disp = ConfusionMatrixDisplay(
        confusion_matrix=mean_cms,
        display_labels=classifier.classes_,
    )
    disp.plot(xticks_rotation="vertical")


def RF_cv(X, y, opt_num_feature="all") -> None:
    selector = SelectKBest(f_classif, k=opt_num_feature)
    X_new = selector.fit_transform(X, y)
    cv = StratifiedKFold(n_splits=10)
    if opt_num_feature == "all":
        classifier = RandomForestClassifier(
            class_weight="balanced",
            max_features="sqrt",
            random_state=42,
        )
    else:
        classifier = RandomForestClassifier(
            n_estimators=200,
            class_weight="balanced",
            max_features="sqrt",
            min_samples_split=2,
            random_state=42,
            max_depth=30,
        )
    tprs = []
    aucs = []
    MCCs = []
    cms = []
    baccs = []
    recalls = []
    precisions = []
    mean_fpr = np.linspace(0, 1, 100)
    fig, ax = plt.subplots()
    for i, (train, test) in enumerate(cv.split(X_new, y)):
        classifier.fit(X_new[train], y[train])
        pred = classifier.predict(X_new[test])
        mcc = matthews_corrcoef(y[test], pred)
        bacc = balanced_accuracy_score(y[test], pred)
        cm = confusion_matrix(y[test], pred, labels=classifier.classes_)
        rec = recall_score(y[test], pred)
        prec = precision_score(y[test], pred)
        viz = RocCurveDisplay.from_estimator(
            classifier,
            X_new[test],
            y[test],
            name=f"ROC fold {i}",
            alpha=0.3,
            lw=1,
            ax=ax,
        )

        interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
        interp_tpr[0] = 0.0
        tprs.append(interp_tpr)
        aucs.append(viz.roc_auc)
        MCCs.append(mcc)
        cms.append(cm)
        baccs.append(bacc)
        recalls.append(rec)
        precisions.append(prec)

    ax.plot(
        [0, 1],
        [0, 1],
        linestyle="--",
        lw=2,
        color="r",
        label="Chance",
        alpha=0.8,
    )
    mean_tpr = np.mean(tprs, axis=0)
    mean_tpr[-1] = 1.0
    mean_auc = auc(mean_fpr, mean_tpr)
    std_auc = np.std(aucs)
    np.mean(MCCs)
    np.std(MCCs)
    stacked_matrices = np.stack(cms, axis=0)
    mean_cms = np.mean(stacked_matrices, axis=0)
    np.mean(baccs)
    np.std(baccs)
    np.mean(recalls)
    np.std(recalls)
    np.mean(precisions)
    np.std(precisions)

    ax.plot(
        mean_fpr,
        mean_tpr,
        color="b",
        label=rf"Mean ROC (AUC = {mean_auc:0.2f} $\pm$ {std_auc:0.2f})",
        lw=2,
        alpha=0.8,
    )
    std_tpr = np.std(tprs, axis=0)
    tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
    tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
    ax.fill_between(
        mean_fpr,
        tprs_lower,
        tprs_upper,
        color="grey",
        alpha=0.2,
        label=r"$\pm$ 1 std. dev.",
    )
    ax.set(
        xlim=[-0.05, 1.05],
        ylim=[-0.05, 1.05],
        title="ROC Curve",
    )
    ax.legend(loc="lower right")
    plt.show()
    disp = ConfusionMatrixDisplay(
        confusion_matrix=mean_cms,
        display_labels=classifier.classes_,
    )
    disp.plot(xticks_rotation="vertical", values_format=".2f")
    plt.show()


def roc_cv_ANOVA_10fold(X, y, k_value, rf=False):
    selector = SelectKBest(f_classif, k=k_value)
    X_new = selector.fit_transform(X, y)
    cv = StratifiedKFold(n_splits=10)
    classifier = MLPClassifier(random_state=42)
    if rf:
        classifier = RandomForestClassifier(
            class_weight="balanced",
            max_features="sqrt",
            random_state=42,
        )
    aucs = []
    MCCs = []
    tprs = []
    mean_fpr = np.linspace(0, 1, 100)
    for _i, (train, test) in enumerate(cv.split(X_new, y)):
        classifier.fit(X_new[train], y[train])
        pred = classifier.predict(X_new[test])
        mcc = matthews_corrcoef(y[test], pred)
        y_score = classifier.predict_proba(X_new[test])[:, 1]
        roc_auc = roc_auc_score(y[test], y_score)
        fpr, tpr, _ = metrics.roc_curve(y[test], y_score)
        interp_tpr = np.interp(mean_fpr, fpr, tpr)
        interp_tpr[0] = 0.0
        tprs.append(interp_tpr)
        aucs.append(roc_auc)
        MCCs.append(mcc)
    return tprs, aucs, MCCs


def feature_selection(X, y, feature_list, name, k_list) -> None:
    MCC_of_different_ANOVA = []
    mean_mcc = []
    mean_fpr = np.linspace(0, 1, 100)
    fig, ax = plt.subplots()
    c = plt.cm.rainbow(np.linspace(0, 1, 20))

    for i, k in enumerate(k_list):
        if k != "all" and int(k) <= len(feature_list):
            tprs, aucs, MCCs = roc_cv_ANOVA_10fold(np.array(X), y, int(k))
        elif k == "all":
            tprs, aucs, MCCs = roc_cv_ANOVA_10fold(np.array(X), y, k)
        else:
            MCC_of_different_ANOVA.append(0)
        mean_tpr = np.mean(tprs, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = auc(mean_fpr, mean_tpr)
        std_auc = np.std(aucs)
        mean_mcc = np.mean(MCCs)

        ax.plot(
            mean_fpr,
            mean_tpr,
            color=c[i],
            label=rf"K={k}, ROC_AUC = {mean_auc:0.2f} $\pm$ {std_auc:0.2f}, MCC = {mean_mcc:0.2f}",
            lw=2,
            alpha=0.4,
        )

    ax.plot(
        [0, 1],
        [0, 1],
        linestyle="--",
        lw=2,
        color="r",
        label="Chance",
        alpha=0.8,
    )

    ax.set(
        xlim=[-0.05, 1.05],
        ylim=[-0.05, 1.05],
        title=rf"ROC Curve ({name})",
    )
    ax.set_xlabel("False Positive Rate (Positive label = 1.0)")
    ax.set_ylabel("True Positive Rate (Positive label = 1.0)")
    ax.legend(loc="lower right", fontsize=7)
    plt.show()


def pairwise_tanimoto_similarity_closest_distance(smi_list, query_set_smi_list, fpe):
    default_value = -1
    results = []
    for my_smi in tqdm(query_set_smi_list, desc="Calculating similarities"):
        sim_results = fpe.similarity(my_smi, 0, n_workers=31)
        sim_dict = dict.fromkeys(smi_list, default_value)
        for idx, value in sim_results:
            p = smi_list[idx]
            sim_dict[p] = max(sim_dict[p], value)
        results.append({"smi": my_smi, "dict": sim_dict})
    temp_save = pd.DataFrame(results)
    temp_save[smi_list] = [list(value_1.values()) for value_1 in temp_save["dict"]]
    temp_save["closest_distance"] = temp_save[smi_list].max(axis=1)
    temp_save["closest_smi"] = temp_save[smi_list].idxmax(axis=1)
    return temp_save.drop(columns=["dict"])


def trialblazer_train(
    training_set,
    y,
    features,
    k,
):
    selector = SelectKBest(f_classif, k=k)
    X = training_set[features]
    X_new = selector.fit_transform(X, y)
    classifier = MLPClassifier(
        hidden_layer_sizes=(10,),
        random_state=42,
        learning_rate_init=0.0001,
        max_iter=600,
    )  # this classifier should be store somewhere so that it won't be trained every time
    classifier.fit(X_new, y)
    return classifier, selector


def trialblazer_func(
    classifier,
    selector,
    test_set,
    threshold,
    training_fpe,
    training_set,
    features,
    remove_MultiComponent_cpd=True,
):
    X_columns = features
    test_set_aligned = test_set.reindex(columns=X_columns, fill_value=0)
    X_test_ANO = selector.transform(test_set_aligned)

    y_prob = classifier.predict_proba(X_test_ANO)
    y_pred_opt = (y_prob[:, 1] >= threshold).astype(int)

    test_set_aligned["id"] = test_set["id"]
    test_set_aligned["SmilesForDropDu"] = test_set["SmilesForDropDu"]
    test_set_aligned["prediction"] = y_pred_opt
    test_set_aligned["pred_prob_positive"] = y_prob[:, 1]
    test_set_aligned["pred_prob_negative"] = y_prob[:, 0]

    # PrOCTOR score
    odd = test_set_aligned.pred_prob_negative / test_set_aligned.pred_prob_positive
    PrOCTOR_score = np.log2(odd)
    test_set_aligned["PrOCTOR_score"] = PrOCTOR_score
    test_set_aligned["PrOCTOR_score"] = test_set_aligned["PrOCTOR_score"].apply(
        lambda x: f"{x:.3f}",
    )
    predict_result_sim = test_set_aligned.iloc[:, -6:]
    predict_result_sim["PrOCTOR_score"] = predict_result_sim["PrOCTOR_score"].astype(
        float,
    )
    if remove_MultiComponent_cpd:
        predict_result_sim_remove_multi = predict_result_sim[
            ~predict_result_sim["id"].str.contains(r"\d+x\d+")
        ]  # Remove multi-component drugs if the toxicity of the compounds is uncertain
        predict_result_sim_remove_multi = predict_result_sim_remove_multi.sort_values(
            by="PrOCTOR_score",
            ascending=False,
            ignore_index=True,
        )
        predict_result_sim = predict_result_sim_remove_multi

    # Applicability domain

    similarity_closest_distance = pairwise_tanimoto_similarity_closest_distance(
        list(training_set["SmilesForDropDu"]),
        list(predict_result_sim["SmilesForDropDu"]),
        training_fpe,
    )
    # prediction output
    predict_result_sim = predict_result_sim.rename(columns={"SmilesForDropDu": "smi"})
    prediction_output = predict_result_sim.merge(
        similarity_closest_distance[["smi", "closest_distance", "closest_smi"]],
        how="left",
        on="smi",
    )
    prediction_output["prediction"] = prediction_output["prediction"].map(
        {0: "benign", 1: "toxic"},
    )

    # return (
    #     predict_result_sim,
    #     similarity_closest_distance[["smi", "closest_distance", "closest_smi"]],
    # )
    return prediction_output


def Trialblazer(
    training_set,
    y,
    features,
    test_set,
    threshold,
    k,
    training_fpe,
    remove_MultiComponent_cpd=True,
    classifier=None,
    selector=None,
):
    if classifier is None:
        classifier, selector = trialblazer_train(
            training_set=training_set, y=y, features=features, k=k,
        )
    return trialblazer_func(
        classifier=classifier,
        selector=selector,
        test_set=test_set,
        threshold=threshold,
        training_fpe=training_fpe,
        remove_MultiComponent_cpd=remove_MultiComponent_cpd,
        features=features,
        training_set=training_set,
    )



import re
import cirpy
import numpy as np
import pandas as pd
import pubchempy as pcp
import spacy
from drug_named_entity_recognition import find_drugs


def ExtractingDrugsNamesAndSmilesMatching(dataset, Name_column):
    FailToExtract, GoToSmilesMatching = Extract_drug_name(dataset, Name_column)
    cirpy_result, GoToPubchempy = cirpy_SmilesMatching(
        FailToExtract,
        GoToSmilesMatching,
    )
    pubchempy_result, GoTo_NextRound = pubchempy_SmilesMatching(
        GoToPubchempy,
        "name_to_use",
    )
    SmilesMatched = pd.concat(
        [cirpy_result, pubchempy_result],
        ignore_index=True,
    ).reset_index()
    return SmilesMatched, GoTo_NextRound


def replace_text(ori_name):
    if pd.isna(ori_name):
        return ori_name
    pattern = [r"(p|P)lus", r"(C|c)ombin", r"(A|a)nd", r"( or )", r"(\+)"]
    ori_name = ori_name.replace("18F", "")
    if ori_name.endswith("mab"):
        return "antibody"
    if any(re.findall(p, ori_name) for p in pattern):
        return "multi-drugs"
    return ori_name


def DNER_with_spacy(DataFrame):
    test_with_spacy = []
    doc_append = []
    nlp = spacy.blank("en")
    for row in DataFrame["replace_Name"].values():
        if pd.notna(row):
            doc = nlp(row)
            doc_append.append(doc)
            with_spacy = find_drugs([t.text for t in doc], is_ignore_case=True)
            test_with_spacy.append(with_spacy)
        else:
            test_with_spacy.append([])
    return test_with_spacy


def extract_results(LM_output):
    name_append = []
    synonyms_append = []
    drugbank_id_append = []
    for row in LM_output:
        if row:
            first_tuple = row[0]
            if isinstance(first_tuple[0], dict):
                name_append.append(first_tuple[0].get("name", np.nan))
                synonyms_append.append(first_tuple[0].get("synonyms", np.nan))
                drugbank_id_append.append(
                    first_tuple[0].get("drugbank_id", np.nan),
                )
            else:
                name_append.append(np.nan)
                synonyms_append.append(np.nan)
                drugbank_id_append.append(np.nan)
        else:
            name_append.append(np.nan)
            synonyms_append.append(np.nan)
            drugbank_id_append.append(np.nan)


    return name_append, synonyms_append, drugbank_id_append


def select_drugs_GoToSecondRound(processed_dataframe):
    s = []
    c = []
    for i, item in enumerate(processed_dataframe["DNER_with_spacy"]):
        if len(item) == 0:
            s.append(i)
        else:
            c.append(i)
    GoToSecondRound = processed_dataframe.iloc[s]
    with_InFo = processed_dataframe.iloc[c]
    return GoToSecondRound, with_InFo


def cirpy_SmilesMatching(FailToExtract, GoToSmilesMatching):
    smiles_append = []
    name = "replace_Name"
    for df in [FailToExtract, GoToSmilesMatching]:
        for row in df[name].values():
            try:
                if row:
                    smiles = cirpy.resolve(row, "smiles")
                    smiles_append.append(smiles)
                else:
                    smiles_append.append(np.nan)
            except Exception:
                smiles_append.append(np.nan)
        name = "name_to_use"
    GoToSmileMatching = pd.concat(
        [FailToExtract, GoToSmilesMatching],
        ignore_index=True,
    )
    GoToSmileMatching["SMILES"] = smiles_append
    cirpy_result = GoToSmileMatching.loc[~GoToSmileMatching["SMILES"].isna()]

    GoToPubchempy = GoToSmileMatching.loc[GoToSmileMatching["SMILES"].isna()]
    if "SMILES_ORI" in GoToPubchempy.columns:
        SMILES_ORI = GoToPubchempy["SMILES_ORI"].to_list()
        GoToPubchempy["SMILES"] = SMILES_ORI

        cirpy_result_2 = GoToPubchempy[GoToPubchempy["SMILES"] != " "]
        GoToPubchempy = GoToPubchempy[GoToPubchempy["SMILES"] == " "]

        cirpy_result = pd.concat(
            [cirpy_result, cirpy_result_2],
            ignore_index=True,
        )
    return cirpy_result, GoToPubchempy


def get_smiles_and_compounds(df, name):
    smiles_append = []
    cs_append = []
    for row in df[name].values():
        if pd.notna(row):
            cs = pcp.get_compounds(row, "name")
            if len(cs) == 1:
                cs_append.append(cs[0])
                smiles_append.append(cs[0].isomeric_smiles)
            elif len(cs) > 1:
                cs_append.append(cs)
                smiles_append.append([item.isomeric_smiles for item in cs])
            else:
                cs_append.append(cs)
                smiles_append.append(np.nan)
        else:
            cs_append.append([])
            smiles_append.append(np.nan)
    return cs_append, smiles_append


def pubchempy_SmilesMatching(GoToPubchempy, name):
    cs_append, smiles_append = get_smiles_and_compounds(GoToPubchempy, name)
    GoToPubchempy["matched_compound"] = cs_append
    GoToPubchempy["SMILES"] = smiles_append

    pubchempy_result_1 = GoToPubchempy.loc[~GoToPubchempy["SMILES"].isna()]
    GoTo_NextRound = GoToPubchempy.loc[GoToPubchempy["SMILES"].isna()]

    name = "replace_Name"
    cs_append, smiles_append = get_smiles_and_compounds(GoTo_NextRound, name)
    GoTo_NextRound["matched_compound"] = cs_append
    GoTo_NextRound["SMILES"] = smiles_append

    pubchempy_result_2 = GoTo_NextRound.loc[~GoTo_NextRound["SMILES"].isna()]
    pubchempy_result = pd.concat([pubchempy_result_1, pubchempy_result_2])
    GoTo_NextRound = GoTo_NextRound.loc[GoTo_NextRound["SMILES"].isna()]
    return pubchempy_result, GoTo_NextRound


def Extract_drug_name(input_df, input_name):
    input_df["replace_Name"] = input_df[input_name].apply(replace_text)
    input_df["DNER_with_spacy"] = DNER_with_spacy(input_df)
    name, synonyms, drugbank_id = extract_results(input_df["DNER_with_spacy"])
    input_df["name_to_use"] = name
    input_df["synonyms"] = synonyms
    input_df["drugbank_id"] = drugbank_id
    name_NeedToExcluded = ["antibody", "multi-drugs", "-"]
    input_df = input_df[
        (input_df["replace_Name"].str.startswith(tuple(name_NeedToExcluded)) is False)
    ]
    FailToExtract, GoToSmilesMatching = select_drugs_GoToSecondRound(input_df)
    return FailToExtract, GoToSmilesMatching



import os
import numpy as np
import pandas as pd
import trialblazer
from trialblazer import Trialblazer

base_model_folder = os.path.join(
    os.path.dirname(trialblazer.__file__),
    "data",
    "base_model",
)

input_file = os.path.join(os.path.dirname(__file__), "data", "test_input.csv")
input_data = pd.read_csv(
    input_file,
    delimiter="|",
)
output_data = pd.read_csv(
    os.path.join(os.path.dirname(__file__), "data", "test_output.csv"),
    delimiter="|",
)

output_data = output_data[["id", "pred_prob_positive", "PrOCTOR_score"]]
output_rm = output_data[~output_data["id"].str.contains(r"\d+x\d+")]

output_data = output_data.set_index("id")
output_data = output_data.sort_index()
output_rm = output_rm.set_index("id")
output_rm = output_rm.sort_index()


def test_run(tmpdir):
    tb = Trialblazer(input_file=input_file)
    tb.run()


def test_download():
    tb = Trialblazer()
    tb.download_model()


def test_run_no_remove(tmpdir):
    tb = Trialblazer(input_file=input_file, remove_MultiComponent_cpd=False)
    tb.run()
    tb.write(output_file=os.path.join(tmpdir, "trialblazer.csv"))
    df = tb.result.copy()
    df = df[["id", "pred_prob_positive", "PrOCTOR_score"]]
    df = df.set_index("id")
    df = df.sort_index()
    assert len(df.index) == len(output_data.index)
    assert np.isclose(df, output_data).all()



from rdkit import Chem
from rdkit.Chem import Crippen
from rdkit.Chem import Descriptors
from rdkit.Chem import Lipinski
from rdkit.Chem import PandasTools
from rdkit.Chem import rdMolDescriptors

"""
Script to calculate physicochemical properties of molecules:
adopted from the script using for ring systems project from Ya Chen's work: https://github.com/anya-chen/RingSystems
"""


def get_physicochemical_properties(molDF, smiles_column) -> None:
    """Applies all property calculations to the ring systems of the dataframe and stores each property in a new column
    :param molDF: dataframe with ring systems as SMILES in the column 'ringSmiles'
    :return: a dataframe with ring system molecules and their properties.
    """
    PandasTools.AddMoleculeColumnToFrame(molDF, smiles_column, "Molecule")
    molDF["N"] = molDF["Molecule"].apply(get_molecule_composition, args=(7,))
    molDF["O"] = molDF["Molecule"].apply(get_molecule_composition, args=(8,))
    molDF["chiral"] = molDF["Molecule"].apply(get_nof_chiral_centers)
    molDF["MW"] = molDF["Molecule"].apply(get_MW)
    molDF["heavy_atoms"] = molDF["Molecule"].apply(num_heavy_atoms)
    molDF["h_acc"] = molDF["Molecule"].apply(
        num_of_h_acceptors_and_donors,
        args=(True,),
    )
    molDF["h_don"] = molDF["Molecule"].apply(
        num_of_h_acceptors_and_donors,
        args=(False,),
    )
    molDF["logP"] = molDF["Molecule"].apply(get_logp)
    molDF["TPSA"] = molDF["Molecule"].apply(get_TPSA)
    molDF["numAro"] = molDF["Molecule"].apply(num_aromatic_atoms)
    molDF["formalCharge"] = molDF["Molecule"].apply(sum_formal_charge)
    molDF["bridgeheadAtoms"] = molDF["Molecule"].apply(num_bridgehead_atoms)
    molDF["frac_csp3"] = molDF["Molecule"].apply(fraction_csp3)


def get_further_physicochemical_properties(molDF) -> None:
    """:param molDF: dataframe with ring systems as SMILES in the column 'ringSmiles'
    :return: a dataframe with ring system molecules and their properties
    """
    del molDF["bridgeheadAtoms"]
    molDF["S"] = molDF["Molecule"].apply(get_molecule_composition, args=(16,))
    molDF["nHalogens"] = molDF["Molecule"].apply(num_halogens)
    molDF["MR"] = molDF["Molecule"].apply(get_mr)


def get_molecule_composition(mol, requestedAtomicNum):
    """Counts the number of atoms of a given element in the ring system
    :param mol: the ring system molecule
    :param requestedAtomicNum: atomic number of the element for which the occurrence should be counted
    :return: the number of atoms of an element.
    """
    counter = 0
    for atom in mol.GetAtoms():
        atomicNum = atom.GetAtomicNum()
        if atomicNum == requestedAtomicNum:
            counter += 1
    return counter


def get_nof_chiral_centers(mol):
    return len(Chem.FindMolChiralCenters(mol, includeUnassigned=True))


def get_MW(mol):
    return round(Descriptors.MolWt(mol), 3)


def num_heavy_atoms(mol):
    return Lipinski.HeavyAtomCount(mol)


def num_of_h_acceptors_and_donors(mol, acc=True):
    if acc:
        return Lipinski.NumHAcceptors(mol)
    return Lipinski.NumHDonors(mol)


def get_logp(mol):
    return round(Crippen.MolLogP(mol), 3)


def get_TPSA(mol):
    return round(Descriptors.TPSA(mol), 3)


def num_aromatic_atoms(mol):
    numAromaticAtoms = 0
    for atom in mol.GetAtoms():
        if atom.GetIsAromatic():
            numAromaticAtoms += 1
    return numAromaticAtoms


def sum_formal_charge(mol):
    formalCharge = 0
    for atom in mol.GetAtoms():
        formalCharge += atom.GetFormalCharge()
    return formalCharge


def num_bridgehead_atoms(mol):
    return rdMolDescriptors.CalcNumBridgeheadAtoms(mol)


def fraction_csp3(mol):
    return round(Descriptors.FractionCSP3(mol), 3)


def num_halogens(mol):
    return Chem.Fragments.fr_halogen(mol)


def get_mr(mol):
    """Wildman-Crippen MR value
    Uses an atom-based scheme based on the values in the paper:
    Wildman and G. M. Crippen JCICS 39 868-873 (1999).
    """
    return round(Crippen.MolMR(mol), 3)



import os
from pathlib import Path
import pandas as pd
from rdkit import Chem

"""
This script is used to remove duplicate molecules for coconut, zinc (datasets have been splitted in parts).
It will:
    - group by unique constitution (same connectivity) and have a constitution ID (conID).
    - within one group, remove smiles with fully undefined stereochemistry if there is a better defined smiles (only consider @ flags in smiles)

use after preprocessing, filter bad patterns.

"""


def remove_duplicate_splitted_files(inputfolder, outputfolder, prefix) -> None:
    """Use for a big file has been splitted in parts.
    Need to put them together first and then split according to consititutions.
    """
    li = []
    for filename in os.listdir(inputfolder):
        file = inputfolder / filename
        df = pd.read_csv(file, sep="\t", index_col=None, header=0)
        li.append(df)

    df = pd.concat(li, axis=0, ignore_index=True)
    df = df[["preprocessedSmiles", "id", "tautomerizedSmiles"]]

    df = df[df["preprocessedSmiles"] != ""]

    # df.drop_duplicates(subset=['preprocessedSmiles'],inplace=True)
    # print('number of entries after drop duplicated preprocessedSmiles: {}'.format(str(len(df))))

    df["nChiralFlags"] = df.apply(count_nFlags_smiles, axis=1)
    df["smiles_noStereo"] = df.apply(get_smiles_noStereo, axis=1)
    df["inchi_noStereo"] = df.apply(get_inchi_nostereo, axis=1)

    df = df.drop_duplicates(subset=["inchi_noStereo"])

    df_noVariance, df_Variance = separate_by_if_stereo_variance(df)

    del df

    outputfilename1 = prefix + "_noVariance_uniq.csv"
    outputFile1 = outputfolder / outputfilename1
    df_noVariance[
        [
            "preprocessedSmiles",
            "id",
            "conID",
            "tautomerizedSmiles",
            "inchi_noStereo",
        ]
    ].to_csv(outputFile1, sep="\t", index=False)
    del df_noVariance

    len(df_Variance)

    outputfilename2 = prefix + "_Variance_uniq.csv"
    outputFile2 = outputfolder / outputfilename2
    if not df_Variance.empty:
        split_dataframe_by_groups_with_tautomerizedSmiles(
            df_Variance,
            outputFile2,
        )


def get_inchi_nostereo(row):
    smi = row["smiles_noStereo"]
    return Chem.MolToInchi(Chem.MolFromSmiles(smi))


def get_smiles_noStereo(row):
    smi = row["preprocessedSmiles"]
    return Chem.MolToSmiles(
        Chem.MolFromSmiles(smi),
        isomericSmiles=False,
    )


def count_nFlags_smiles(row):
    # annotate how many chiral tags in the smiles
    smi = row["preprocessedSmiles"]
    smi = smi.replace("@@", "@")
    return smi.count("@")


def separate_by_if_stereo_variance(df):
    df = df.groupby("smiles_noStereo")

    frames = []
    frames_variance = []
    conID = 0
    for _name, group in df:
        group.count().reset_index()
        conID += 1
        group["conID"] = (
            conID  # can not compare this ID of compounds in different datasets.
        )

        if len(group) == 1:
            frames.append(group)
        else:
            frames_variance.append(group)

    df_noVariance = pd.concat(frames)

    if frames_variance:
        df_Variance = pd.concat(frames_variance)
    else:
        df_Variance = pd.DataFrame()

    return df_noVariance, df_Variance


def split_dataframe_by_groups_with_tautomerizedSmiles(
    df,
    outputFile,
    size=5000,
) -> None:
    """Use this to split a set to smaller parts."""
    df = df.groupby("smiles_noStereo")

    frames = []
    i = 0
    for _name, group in df:
        group.count().reset_index()
        frames.append(group)
        if len(frames) % size == 0:
            i += 1
            results_df = remove_smiles_fully_undefined_stereo(
                pd.concat(frames),
            )
            outputFile3 = str(outputFile).split(".")[0] + "_part" + str(i) + ".csv"
            results_df[
                [
                    "preprocessedSmiles",
                    "id",
                    "conID",
                    "tautomerizedSmiles",
                    "inchi_noStereo",
                ]
            ].to_csv(outputFile3, sep="\t", index=False)
            del results_df
            frames = []
    i += 1
    results_df = remove_smiles_fully_undefined_stereo(pd.concat(frames))
    outputFile3 = str(outputFile).split(".")[0] + "_part" + str(i) + ".csv"
    results_df[
        [
            "preprocessedSmiles",
            "id",
            "conID",
            "tautomerizedSmiles",
            "inchi_noStereo",
        ]
    ].to_csv(outputFile3, sep="\t", index=False)
    del results_df


def remove_smiles_fully_undefined_stereo(df):
    """Only when there is other smiles in this consititution."""
    df = df.groupby("smiles_noStereo")

    frames = []
    for _name, group in df:
        group.count().reset_index()
        nums = set(group.nChiralFlags)

        if len(nums) > 1 and (0 in nums):
            new_group = group[group.nChiralFlags != 0]
        else:
            new_group = group

        new_group.count().reset_index()
        frames.append(new_group)

    return pd.concat(frames)


if __name__ == "__main__":
    refinedInputFolder = Path(
        "/home/ychen/projects/nps_ringsys/20210618_preprocessing/refinedCoconut/noBadPatternSmiles",
    )
    refinedOutputFolder = refinedInputFolder.parent / "uniqueSmiles"
    if not refinedOutputFolder.exists():
        refinedOutputFolder.mkdir()

    remove_duplicate_splitted_files(
        refinedInputFolder,
        refinedOutputFolder,
        "refinedCoconut",
    )

    removedInputFolder = Path(
        "/home/ychen/projects/nps_ringsys/20210618_preprocessing/removedCoconut/noBadPatternSmiles",
    )
    removedOutputFolder = removedInputFolder.parent / "uniqueSmiles"
    if not removedOutputFolder.exists():
        removedOutputFolder.mkdir()

    remove_duplicate_splitted_files(
        removedInputFolder,
        removedOutputFolder,
        "removedCoconut",
    )

    zincInputFolder = Path(
        "/home/ychen/projects/nps_ringsys/20210519_data_prep/ZINC/noBadPatternSmiles",
    )
    zincOutputFolder = zincInputFolder.parent / "uniqueSmiles"
    if not zincOutputFolder.exists():
        zincOutputFolder.mkdir()

    remove_duplicate_splitted_files(zincInputFolder, zincOutputFolder, "zinc")




import os
from pathlib import Path
from typing import Optional
import numpy as np
import pandas as pd
from tqdm import tqdm


def separate_similarity_results(
    results,
    target_id_list,
    preprocessed_target_unique_smiles,
    output_path_temp_save,
    training_data=False,
    start_index=0,
) -> None:
    default_value = -1
    for i, array in tqdm(
        generator_separate_results(results, start_index=start_index),
    ):
        smi = results.my_smi.iloc[i]
        temp_save = pd.DataFrame()
        dict_save = []
        smi_save = []
        sim_dict = dict.fromkeys(target_id_list, default_value)
        for value in array:
            if training_data:
                p = preprocessed_target_unique_smiles.iloc[value[0]]["target_id"]
            else:
                p = preprocessed_target_unique_smiles.iloc[value[0]]["target_id"]
            if any(item in target_id_list for item in p):
                for target in p:
                    sim_dict[target] = max(sim_dict[target], value[1])
        dict_save.append(sim_dict)
        smi_save.append(smi)
        temp_save["smi"] = smi_save
        temp_save["dict"] = dict_save
        temp_save.to_csv(
            Path(output_path_temp_save.name) / f"temp_save_{i}.csv",
            sep="|",
        )


def organize_similarity_results(
    target_id_list,
    output_path_temp_save,
    training_data=False,
):
    whole_df = read_results_from_file(output_path_temp_save)
    parsed_dicts = []
    for value in tqdm(whole_df["dict"]):
        parsed = eval(value, {"__builtins__": {}, "nan": float("nan"), "np": np})
        parsed_dicts.append(parsed)
    whole_df["dict_con"] = parsed_dicts
    all_lists = []
    target_cols_found = set()
    for value_dict in whole_df["dict_con"]:
        if training_data:
            values = list(value_dict.values())
            all_lists.append(values)
        else:
            values = []
            for target in target_id_list:
                if target in value_dict:
                    target_cols_found.add(target)
                    values.append(value_dict[target])
            all_lists.append(values)
    if training_data:
        whole_df[target_id_list] = all_lists
        return whole_df, target_id_list
    whole_df[list(target_cols_found)] = all_lists
    return whole_df, list(target_cols_found)


def read_results_from_file(output_temp_save):
    whole_df = pd.DataFrame()
    for file in tqdm(os.listdir(output_temp_save.name)):
        filename = Path(output_temp_save.name) / file
        df1 = pd.read_csv(filename, sep="|")
        whole_df = pd.concat([whole_df, df1], ignore_index=True)
    return whole_df


def binarize_similarity_value(dataframe, target_list, threshold):
    dataframe[target_list] = dataframe[target_list].apply(
        lambda row: row.map(lambda x: threshold_binarize(x, threshold)),
    )
    return target_list, dataframe


def remove_tested_inactive_targets(
    binarized_inactive_target,
    binarized_query_targets,
):
    filtered_rows = []
    for i, targets in binarized_inactive_target.iterrows():
        inactive_targets = set(targets[targets == "1"].index)
        active_targets = binarized_query_targets.iloc[i, :]
        active_target_indices = set(
            active_targets[active_targets == "1"].index,
        )
        common_targets = inactive_targets & active_target_indices
        if common_targets:
            active_targets[list(common_targets)] = "0"
        filtered_rows.append(active_targets)
    filtered_df = pd.DataFrame(filtered_rows)
    filtered_df["SmilesForDropDu"] = binarized_query_targets["SmilesForDropDu"]
    target_list = filtered_df.filter(regex="^(CHEMBL*)").columns.tolist()
    return filtered_df, target_list


def remove_invariant_target(sanity_checked_dataframe, target_list):
    unique_counts = sanity_checked_dataframe[target_list].nunique()
    unique_counts[unique_counts == 1]
    target_binarize_list = unique_counts[unique_counts != 1].index.to_list()
    binarized_target_remain = sanity_checked_dataframe[target_binarize_list]
    binarized_target_remain["SmilesForDropDu"] = sanity_checked_dataframe[
        "SmilesForDropDu"
    ]
    return binarized_target_remain, target_binarize_list


def threshold_binarize(x, threshold) -> Optional[str]:
    if x >= threshold:
        return "1"
    if x < threshold:
        return "0"
    return None


def generator_separate_results(results, start_index=0):
    yield from enumerate(
        results.similarity_results[start_index:],
        start=start_index,
    )



import base64
import hashlib
import multiprocessing as mp
import os
import pandas as pd
from .preprocess_molecules import preprocess_database
from .split_files import split_large_files


# def get_data_from_DB(inputFile, reformatedFile):
# smilesDict = dict()
# with open(inputFile, encoding="utf-8") as FoodDBFile:
#     moleculeCsv = pd.read_csv(FoodDBFile, delimiter=None)
def get_data_from_DB(moleculeCsv, reformatedFile, smiles_col="SMILES", id_col=None):
    smilesDict = {}
    smiles = moleculeCsv[smiles_col]
    moleculeCsv["hash_id"] = moleculeCsv[smiles_col].apply(
        lambda x: base64.urlsafe_b64encode(hashlib.md5(x.encode("utf8")).digest())
        .decode()
        .strip()
        .strip("="),
    )
    if id_col is not None:
        ID = moleculeCsv[id_col]
    elif "chembl_id" in moleculeCsv:
        ID = moleculeCsv["chembl_id"]
    else:
        moleculeCsv["chembl_id"] = None
        ID = moleculeCsv["hash_id"]
    smilesDict = dict(zip(smiles, ID))
    with open(reformatedFile, "w", encoding="utf-8") as f:
        pd.DataFrame.from_dict(smilesDict, orient="index").to_csv(f)
    return smilesDict


def get_molecule_dict_as_smi_output(smilesDict, outputFile) -> None:
    with open(outputFile, "w", encoding="utf-8") as of:
        of.write("smiles ID\n")
        for smiles, ID in smilesDict.items():
            if smiles != "":
                of.write(f"{smiles} {ID}\n")


def preprocess(moleculeCsv, outputFolder, smiles_col="SMILES", id_col=None) -> None:
    # def preprocess(inputFile, outputFolder):
    if not outputFolder.exists():
        outputFolder.mkdir()

    reformatedFolder = outputFolder / "smiles_reformat"
    if not reformatedFolder.exists():
        reformatedFolder.mkdir()
    reformatedName = "trialblazer_smiles.smi"
    # reformatedName = inputFile.split("/")[-1].split(".")[0] + ".smi"
    reformatedFile = reformatedFolder / reformatedName

    smilesDict = get_data_from_DB(
        moleculeCsv, reformatedFile, smiles_col=smiles_col, id_col=id_col,
    )
    # smilesDict = get_data_from_DB(inputFile, reformatedFile)
    get_molecule_dict_as_smi_output(smilesDict, reformatedFile)

    # Split it into multiple files
    splitFolder = outputFolder / "smiles_input"
    if not splitFolder.exists():
        splitFolder.mkdir()

    split_large_files(reformatedFile, 2000)

    # structure preprocess
    preprocessedDir = outputFolder / "preprocessedSmiles"
    if not preprocessedDir.exists():
        preprocessedDir.mkdir()

    databaseFilesJobList = []
    for subfile in os.listdir(splitFolder):
        databaseFilesJobList.append(os.path.join(splitFolder, subfile))
    pool = mp.Pool(processes=max(1, mp.cpu_count() - 1))
    pool.map(preprocess_database, databaseFilesJobList)
    pool.close()
    pool.join()


def CheckOutputResult(outputFolder) -> None:
    logdir = outputFolder / "log"
    if not logdir.exists():
        logdir.mkdir()
    CheckErrorFile = outputFolder / "error.csv"
    for filename in os.listdir(logdir):
        if filename.endswith(".log"):
            log_files = os.path.join(logdir, filename)
        with open(log_files) as f:
            lines = f.readlines()
            CheckFile = {}
            for line in lines:
                line = line.strip()
                Smiles = line.split(" ", 1)[0]
                error = line.split(" ", 1)[1]
                CheckFile[Smiles] = error
        with open(CheckErrorFile, "a") as of:
            for key, value in CheckFile.items():
                if value != "[]":
                    of.write(f"{key}:\t{value}\n")



import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib import cm


def plot_score_distribution_with_significance(data, p_values) -> None:
    custom_palette = {
        "Benign compounds\nTraining set": "#ff7f00",
        "Toxic compounds\nTraining set": "#984ea3",
        "Benign compounds\nTest set": "orangered",
    }
    plt.figure(figsize=(6, 6))
    sns.violinplot(x="Category", y="Value", data=data, palette=custom_palette, inner="quartile")
    y_max = data["Value"].max() + 0.5
    step = 5
    for i, ((cat1, cat2), p) in enumerate(p_values.items()):
        x1, x2 = list(data["Category"].unique()).index(cat1), list(data["Category"].unique()).index(cat2)
        y = y_max + ((i + 1) * step)
        plt.plot([x1, x1, x2, x2], [y, y + 1, y + 1, y], lw=1.5, color="black")
        p_text = f"p = {p:.2e}"
        plt.text((x1 + x2) / 2, y + 2, p_text, ha="center", fontsize=11, fontweight="bold")
    plt.ylim(-15, 70)
    plt.xlabel("Category", labelpad=10)
    plt.ylabel("PrOCTOR Score")
    plt.show()

def plot_correlation(predict_result_toxic, predict_result_benign) -> None:
    plt.figure(figsize=(10, 6))
    all_logP = predict_result_toxic["logP"].tolist() + predict_result_benign["logP"].tolist()
    all_scores = predict_result_toxic["PrOCTOR_score"].tolist() + predict_result_benign["PrOCTOR_score"].tolist()

    correlation_logP = np.corrcoef(all_logP, all_scores)[0, 1]

    scatter = plt.scatter(predict_result_toxic["mw"], predict_result_toxic["PrOCTOR_score"],
                        c=predict_result_toxic["logP"], cmap="plasma", alpha=0.9)
    plt.scatter(predict_result_benign["mw"], predict_result_benign["PrOCTOR_score"],
                c=predict_result_benign["logP"], cmap="plasma", alpha=0.9)

    cbar = plt.colorbar(scatter)
    cbar.set_label("logP")

    all_mw = predict_result_toxic["mw"].tolist() + predict_result_benign["mw"].tolist()
    correlation_mw = np.corrcoef(all_mw, all_scores)[0, 1]

    decision_threshold = 0.06264154114736771
    prediction_threshold = np.log2((1 - decision_threshold) / decision_threshold)
    plt.axhline(y=prediction_threshold, color="navy", linestyle="--", linewidth=1.5, label="Prediction threshold")

    plt.text(633, 2.2, f"Threshold = {prediction_threshold:.2f}", fontsize=10, color="black")
    plt.text(140, 32, f"r (MW) = {correlation_mw:.2f}", fontsize=10, color="black")
    plt.text(140, 30.5, f"r (logP) = {correlation_logP:.2f}", fontsize=10, color="black")

    plt.xlabel("Molecular Weight (mw)")
    plt.ylabel("PrOCTOR Score")
    plt.legend()
    plt.grid(True)
    plt.show()

def SuspectedAdverseDrugEvents_count_for_eachdrug(prediction_combine) -> None:
    num_categories = len(prediction_combine.columns[3:])
    color_map = cm.get_cmap("tab20b", num_categories)
    colors = [mcolors.rgb2hex(color_map(i)) for i in range(num_categories)]

    bar_width = 0.8
    gap_between_drugs = 25

    x_indexes = np.arange(len(prediction_combine["Drugs"])) * (1 + gap_between_drugs)

    plt.figure(figsize=(30, 10))
    for i, (column, color) in enumerate(zip(prediction_combine.columns[3:], colors)):
        plt.bar(x_indexes + i * bar_width, prediction_combine[column], width=bar_width,
                label=column, color=color)

    xtick_positions = x_indexes + bar_width * (num_categories - 1) / 2
    plt.xticks(ticks=xtick_positions, labels=prediction_combine["Drugs"],rotation=45, ha="right", fontsize=8)

    plt.xlabel("Drug", fontsize=10)
    plt.ylabel("Number of adverse drug reaction reports", fontsize=10)

    mid_index = len(prediction_combine["Drugs"]) // 2
    separation_x = (x_indexes[mid_index - 1] + x_indexes[mid_index + 1]) / 2
    plt.axvline(x=separation_x, color="navy", linestyle="--", linewidth=2)

    plt.text(separation_x - 10, plt.ylim()[1] * 0.9, "Benign", fontsize=16,fontweight="bold", ha="right", color="#8ea5c8")
    plt.text(separation_x + 10, plt.ylim()[1] * 0.9, "Toxic", fontsize=16,fontweight="bold", ha="left", color="#a17db4")
    plt.legend(title="Severity", loc="upper center", bbox_to_anchor=(0.5, -0.25), ncol=5)
    plt.tight_layout()
    plt.show()

def SuspectedAdverseDrugEvents_count(pre_benign, pre_toxic, list_of_adverse_reaction) -> None:
    benign_not_serious_list_array, benign_serious_list_array = process_adverse_reactions(pre_benign, list_of_adverse_reaction)
    toxic_not_serious_list_array, toxic_serious_list_array = process_adverse_reactions(pre_toxic, list_of_adverse_reaction)
    category_counts_benign = {
        "Non Serious": benign_not_serious_list_array.sum(axis=1),
        "Serious": benign_serious_list_array.sum(axis=1),
    }
    category_counts_toxic = {
        "Non Serious": toxic_not_serious_list_array.sum(axis=1),
        "Serious": toxic_serious_list_array.sum(axis=1),
    }
    width = 0.4
    base_color_benign = "#ff7f00"
    base_color_toxic = "#984ea3"
    alphas = [0.6,0.8]
    offset = 0.2

    fig, ax = plt.subplots(figsize=(12, 6))
    bottom_benign = np.zeros(len(list_of_adverse_reaction))
    bottom_toxic = np.zeros(len(list_of_adverse_reaction))
    x_positions = np.arange(len(list_of_adverse_reaction)) * 1.2

    for i, (label, counts) in enumerate(category_counts_benign.items()):
        ax.bar(
            [x - offset for x in x_positions], counts, width=width, label=f"Predicted benign: {label}", color=base_color_benign,
            alpha=alphas[i], edgecolor="black", linewidth=1, bottom=bottom_benign,
        )
        bottom_benign += counts

    for i, (label, counts) in enumerate(category_counts_toxic.items()):
        ax.bar(
            [x + offset for x in x_positions], counts, width=width, label=f"Predicted toxic: {label}",
            color=base_color_toxic, alpha=alphas[i], edgecolor="black", linewidth=1, bottom=bottom_toxic,
        )
        bottom_toxic += counts

    ax.set_xticks(x_positions)
    ax.set_xticklabels(list_of_adverse_reaction, rotation=45, ha="right", fontsize=10)
    ax.set_ylabel("Count", fontsize=12)
    ax.set_xlabel("Adverse drug reaction", fontsize=12)
    ax.legend()
    plt.show()

def SuspectedAdverseDrugEvents_Totalcount_for_eachdrug(prediction_combine) -> None:
    x_indexes = np.arange(len(prediction_combine["Drugs"]))
    fig, ax = plt.subplots(figsize=(12, 6))
    plt.bar(x_indexes, prediction_combine["Total"], color="orangered")
    mid_index = len(prediction_combine["Drugs"]) // 2
    separation_x = (x_indexes[mid_index - 1] + x_indexes[mid_index]) / 2
    plt.axvline(x=separation_x, color="navy", linestyle="--", linewidth=2)
    plt.text(separation_x - 3, plt.ylim()[1] * 0.9, "Benign", fontsize=10, fontweight="bold",ha="right", color="#8ea5c8")
    plt.text(separation_x + 3, plt.ylim()[1] * 0.9, "Toxic", fontsize=10, fontweight="bold",ha="left", color="#a17db4")
    plt.xticks(ticks=x_indexes, labels=prediction_combine["Drugs"], rotation=45, ha="right",fontsize=8)
    plt.xlabel("Drug")
    plt.ylabel("Total count of adverse reactions")
    plt.tight_layout()
    plt.show()

def process_adverse_reactions(prediction, list_of_adverse_reaction):
    not_serious_list, serious_list = [], []
    for reaction in list_of_adverse_reaction:
        not_serious_list.append(prediction.loc[prediction["Seriousness"] == "Non Serious", reaction].tolist())
        serious_list.append(prediction.loc[prediction["Seriousness"] == "Serious", reaction].tolist())
    not_serious_list_array = np.array(not_serious_list)
    serious_list_array = np.array(serious_list)
    return not_serious_list_array, serious_list_array

